from typing import List

import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.line import Line
from sqlfmt.mode import Mode
from sqlfmt.node_manager import NodeManager
from sqlfmt.splitter import LineSplitter
from tests.util import read_test_data


@pytest.fixture
def splitter(default_mode: Mode) -> LineSplitter:
    node_manager = NodeManager(default_mode.dialect.case_sensitive_names)
    return LineSplitter(node_manager)


@pytest.fixture
def default_analyzer(default_mode: Mode) -> Analyzer:
    return default_mode.dialect.initialize_analyzer(default_mode.line_length)


@pytest.fixture
def depth_split_line(default_analyzer: Analyzer) -> Line:
    source_string = "with my_cte as (select 1, b from my_schema.my_table),\n"
    raw_query = default_analyzer.parse_query(source_string)
    return raw_query.lines[0]


def test_maybe_split(splitter: LineSplitter, depth_split_line: Line) -> None:
    line_gen = splitter.maybe_split(depth_split_line)
    result = list(map(str, line_gen))

    expected = [
        "with\n",
        " " * 4 + "my_cte\n",
        " " * 4 + "as (\n",
        " " * 8 + "select\n",
        " " * 12 + "1,\n",
        " " * 12 + "b\n",
        " " * 8 + "from\n",
        " " * 12 + "my_schema.my_table\n",
        " " * 4 + "),\n",
    ]

    assert result == expected


def test_split_one_liner(splitter: LineSplitter, default_analyzer: Analyzer) -> None:
    source_string = "select * from my_table\n"
    raw_query = default_analyzer.parse_query(source_string)

    for raw_line in raw_query.lines:
        splits = splitter.maybe_split(raw_line)
        result = list(splits)
        assert len(result) == 4


def test_simple_comment_split(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string, expected_result = read_test_data(
        "unit_tests/test_splitter/test_simple_comment_split.sql"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = "".join([str(line) for line in split_lines])

    assert actual_result == expected_result

    expected_comments = [
        (
            "-- not distinct, just an ordinary select here,"
            " no big deal at all, it's okay really"
        ),
        "",
        "-- here is a long comment to be wrapped above this line",
        "-- a short comment",
        "-- standalone for third",
        "-- here is another long comment to be wrapped but not indented",
        "",
        "-- another comment that is a little bit too long to stay here",
        "-- this should stay",
        "",
        "-- one last comment",
    ]

    for i, line in enumerate(split_lines):
        if line.comments:
            assert str(line.comments[0]) == expected_comments[i]


def test_split_count_window_function(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = (
        "count(case when a is null then 1 end) over "
        "(partition by user_id, date_trunc('year', performed_at)) as d,\n"
    )
    expected_result = (
        "count(\n"
        "    case\n"
        "        when\n"
        "            a\n"
        "            is null\n"
        "        then\n"
        "            1\n"
        "    end\n"
        ")\n"
        "over (\n"
        "    partition by\n"
        "        user_id,\n"
        "        date_trunc(\n"
        "            'year',\n"
        "            performed_at\n"
        "        )\n"
        ")\n"
        "as d,\n"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = "".join([str(line) for line in split_lines])

    assert actual_result == expected_result


def test_split_function_no_args(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = "sum(a) over () as a,\nrow_number() over () as b,\n"
    expected_result = [
        "sum(\n",
        "    a\n",
        ")\n",
        "over (\n",
        ")\n",
        "as a,\n",
        "row_number(\n",
        ")\n",
        "over (\n",
        ")\n",
        "as b,\n",
    ]
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]

    assert actual_result == expected_result


def test_comment_split_impact_on_open_brackets(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string, expected_result = read_test_data(
        "unit_tests/test_splitter/test_comment_split_impact_on_open_brackets.sql"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = "".join([str(line) for line in split_lines])
    assert actual_result == expected_result


def test_split_long_line_on_operator(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = (
        "a_really_long_field + a_really_really_really_long_field "
        "+ a_really_really_really_really_long_field as another_field\n"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "a_really_long_field\n",
        "+ a_really_really_really_long_field\n",
        "+ a_really_really_really_really_long_field\n",
        "as another_field\n",
    ]
    assert actual_result == expected_result


def test_split_at_star(splitter: LineSplitter, default_analyzer: Analyzer) -> None:
    source_string = "select *, my_table.*, 1 * 1, a_field * b_field from my_table\n"
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "select\n",
        "    *,\n",
        "    my_table.*,\n",
        "    1\n",
        "    * 1,\n",
        "    a_field\n",
        "    * b_field\n",
        "from\n",
        "    my_table\n",
    ]
    assert actual_result == expected_result


def test_jinja_block_split(splitter: LineSplitter, default_analyzer: Analyzer) -> None:
    source_string, expected_result = read_test_data(
        "unit_tests/test_splitter/test_jinja_block_split.sql"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = "".join([str(line).lstrip() for line in split_lines])

    assert actual_result == expected_result


def test_split_at_and(splitter: LineSplitter, default_analyzer: Analyzer) -> None:
    source_string = "select 1 where a between b and c and d between e and f and a < b\n"
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "select\n",
        "    1\n",
        "where\n",
        "    a\n",
        "    between b\n",
        "    and c\n",
        "    and d\n",
        "    between e\n",
        "    and f\n",
        "    and a\n",
        "    < b\n",
    ]
    assert actual_result == expected_result


def test_split_before_semicolon(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = "select 1; select 2; select 3; select 4;\n"
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "select\n",
        "    1\n",
        ";\n",
        "select\n",
        "    2\n",
        ";\n",
        "select\n",
        "    3\n",
        ";\n",
        "select\n",
        "    4\n",
        ";\n",
    ]
    assert actual_result == expected_result


def test_split_around_set_operator(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = (
        "select 1 union all select 2 union all (select 3) union all select 4\n"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "select\n",
        "    1\n",
        "union all\n",
        "select\n",
        "    2\n",
        "union all\n",
        "(\n",
        "    select\n",
        "        3\n",
        ")\n",
        "union all\n",
        "select\n",
        "    4\n",
    ]
    assert actual_result == expected_result


def test_split_between_brackets(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = "select split(my_string, ',')[offset(0)], my_dict['tlk']('arg')\n"
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    expected_result = [
        "select\n",
        "    split(\n",
        "        my_string,\n",
        "        ','\n",
        "    )\n",
        "    [\n",
        "        offset(\n",
        "            0\n",
        "        )\n",
        "    ],\n",
        "    my_dict\n",
        "    [\n",
        "        'tlk'\n",
        "    ]\n",
        "    (\n",
        "        'arg'\n",
        "    )\n",
    ]
    assert actual_result == expected_result


def test_split_around_data(splitter: LineSplitter, default_analyzer: Analyzer) -> None:
    source_string = "{% set foo %}//my one-line //data{% endset %}\n"
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [str(line) for line in split_lines]
    assert actual_result == [source_string]


def test_can_split_very_long_lines(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_splitter/test_very_long_single_line.sql"
    )
    raw_query = default_analyzer.parse_query(source_string)

    split_lines: List[Line] = []

    # just make sure this doesn't raise a maximum recursion depth error
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))


def test_maybe_split_no_terminating_newline(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = "select 1, 2, 3\n"

    raw_query = default_analyzer.parse_query(source_string)

    assert len(raw_query.lines) == 1
    new_line = Line.from_nodes(
        previous_node=None, nodes=raw_query.lines[0].nodes[:-1], comments=[]
    )

    assert not new_line.nodes[-1].is_newline
    split_lines = splitter.maybe_split(new_line)

    actual_result = [str(line) for line in split_lines]

    expected_result = [
        "select\n",
        "    1,\n",
        "    2,\n",
        "    3\n",
    ]

    assert actual_result == expected_result


def test_split_leading_comma_comment(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = """
    select 1
        -- two
        , 2  -- two inline
        -- three
        , 3  -- three inline
    """.strip()

    raw_query = default_analyzer.parse_query(source_string)
    assert len(raw_query.lines) == 3

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [line.render_with_comments(88) for line in split_lines]
    expected_result = [
        "select\n",
        "    1\n",
        "    ,\n",
        "    -- two\n    2  -- two inline\n",
        "    ,\n",
        "    -- three\n    3  -- three inline\n",
    ]
    assert actual_result == expected_result


def test_split_trailing_operator_comment(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = """
    select
        1 + -- one
        2 + -- two
        3 + 4 -- four
    """.strip()

    raw_query = default_analyzer.parse_query(source_string)
    assert len(raw_query.lines) == 4

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    actual_result = [line.render_with_comments(88) for line in split_lines]
    expected_result = [
        "select\n",
        "    1  -- one\n",
        "    +\n",
        "    2  -- two\n",
        "    +\n",
        "    3\n",
        "    + 4  -- four\n",
    ]
    assert actual_result == expected_result


def test_split_formatting_disabled(
    splitter: LineSplitter, default_analyzer: Analyzer
) -> None:
    source_string = """
    -- fmt: off
    select 1, 2, 3
    """.strip()

    raw_query = default_analyzer.parse_query(source_string)
    assert len(raw_query.lines) == 2
    select_line = raw_query.lines[1]
    assert select_line.formatting_disabled
    assert select_line.nodes[0].is_unterm_keyword
    assert select_line.nodes[0].formatting_disabled

    split_lines: List[Line] = []
    for raw_line in raw_query.lines:
        split_lines.extend(splitter.maybe_split(raw_line))

    assert split_lines == raw_query.lines



import ast
import keyword
import re
from dataclasses import dataclass, field
from importlib import import_module
from itertools import chain, product
from types import ModuleType
from typing import Dict, List, MutableSet, NamedTuple, Optional, Tuple

from sqlfmt.line import Line
from sqlfmt.mode import Mode
from sqlfmt.node import Node
from sqlfmt.node_manager import NodeManager
from sqlfmt.splitter import LineSplitter


class BlackWrapper:
    """
    A thin wrapper around black. Tries to import black when
    instantiated. Provides a safe interface, format_string
    """

    PY_RESERVED_WORDS = list(keyword.kwlist)

    class StringProperties(NamedTuple):
        has_newlines: bool
        keyword_replacements: Dict[str, int]
        tilde_replacements: Dict[str, int]

    def __init__(self) -> None:
        try:
            self.black: Optional[ModuleType] = import_module("black")
        except ImportError:
            self.black = None

    def format_string(self, source_string: str, max_length: int) -> Tuple[str, bool]:
        """
        Attempt to use black to format source_string to a line_length of max_length.
        Return source_string if black isn't installed or it can't parse source_string.

        Return a tuple of the formatted string and a boolean that indicates whether
        black successfully ran on the string
        """
        BLACKENED = True
        NOT_BLACKENED = False

        if not self.black:
            return source_string, NOT_BLACKENED

        try:
            preprocessed_string, string_properties = self._preprocess_string(
                source_string
            )
        except ValueError:
            return source_string, NOT_BLACKENED

        black_mode = self.black.Mode(line_length=max_length)

        try:
            formatted_string = self.black.format_str(
                preprocessed_string, mode=black_mode
            ).rstrip()
        except ValueError:
            # the string isn't valid python.
            # Jinja allows linebreaks where python doesn't
            # so let's try again without newlines in the code
            if string_properties.has_newlines:
                try:
                    flat_code = preprocessed_string.replace("\n", " ")
                    formatted_string = self.black.format_str(
                        flat_code, mode=black_mode
                    ).rstrip()
                except ValueError:
                    # there is other jinja syntax that isn't valid python,
                    # so if this still fails, just stop trying
                    return source_string, NOT_BLACKENED
                else:
                    postprocessed_string = self._postprocess_string(
                        formatted_string, string_properties
                    )
                    return postprocessed_string, BLACKENED
            else:
                return source_string, NOT_BLACKENED
        else:
            postprocessed_string = self._postprocess_string(
                formatted_string, string_properties
            )
            return postprocessed_string, BLACKENED

    @classmethod
    def _preprocess_string(cls, source_string: str) -> Tuple[str, StringProperties]:
        """
        Takes a jinja source_string and performs some small transformations on it to
        make it valid python that black can format:
        1. Detects newlines
        2. substitutes python keywords like return for safe alternatives like return_
        3. substitutes jinja tilde operators with another binary operator

        Returns a tuple of the processed string and a NamedTuple of the stats from
        pre-processing
        """
        has_newline = True if "\n" in source_string else False

        processed_string, keyword_replacements = cls._replace_reserved_words(
            source_string=source_string
        )
        processed_string, tilde_replacements = cls._replace_tildes(
            source_string=processed_string
        )

        props = cls.StringProperties(
            has_newlines=has_newline,
            keyword_replacements=keyword_replacements,
            tilde_replacements=tilde_replacements,
        )
        return processed_string, props

    @classmethod
    def _replace_reserved_words(cls, source_string: str) -> Tuple[str, Dict[str, int]]:
        """
        Replaces python reserved words in source_string when they are used as variables
        or function names. Returns a string with the replacements made and a dict of
        the number and types of replacements made
        """
        suffixes = [r"\s*=", r"\("]
        replacements = {
            # kw patt: replacement patt, replacement
            # e.g. r"return\(": (r"return_\(", "return_(")
            f"{w}{s}": (f"{w}_{s}", f"{w}_{s[-1]}")
            for (w, s) in product(cls.PY_RESERVED_WORDS, suffixes)
        }

        # check to make sure there aren't already variables with the replacement
        # names in the source string
        preexisting_sentinels = any(
            [
                re.search(repl_patt, source_string)
                and re.search(kw_patt, source_string)
                for kw_patt, (repl_patt, _) in replacements.items()
            ]
        )
        if preexisting_sentinels:
            # abort
            raise ValueError("Cannot preprocess due to preexisting sentinels")

        # try to replace any instances of reserved words with a safe alternative
        processed_string = source_string
        keyword_replacements = {}
        for patt, (repl_patt, repl) in replacements.items():
            processed_string, n = re.subn(patt, repl, processed_string)
            if n > 0:
                keyword_replacements[repl_patt] = n

        return processed_string, keyword_replacements

    @classmethod
    def _replace_tildes(cls, source_string: str) -> Tuple[str, Dict[str, int]]:
        """
        Jinja uses ~ as the string concatenation operator, but black cannot parse the
        tilde. This method finds another operator to safely replace the tilde with,
        and returns a string with the tilde replaced and a dict with the symbol it
        was replaced with and the number of replacements made.
        """
        if "~" in source_string:
            operators = ["+", "-", "*", "/"]
            for operator in operators:
                if operator in source_string:
                    continue
                else:
                    n = source_string.count("~")
                    processed_string = source_string.replace("~", operator)
                    return processed_string, {operator: n}
            else:
                return source_string, {}
        else:
            return source_string, {}

    @classmethod
    def _postprocess_string(
        cls,
        formatted_string: str,
        string_properties: StringProperties,
    ) -> str:
        """
        Translates a formatted python string back to jinja. Undoes some pre-processing
        """

        def remove_underscore(m: re.Match) -> str:
            s: str = m.group(0)
            # All matches should only have a single underscore
            assert s.count("_") == 1, "Internal Error! Please open an issue"
            return s.replace("_", "")

        for repl_patt, n in string_properties.keyword_replacements.items():
            formatted_string, k = re.subn(
                repl_patt, remove_underscore, formatted_string
            )
            assert n == k, (
                "Internal Error! Did not reverse the same number of keywords that "
                "were replaced. Please open an issue"
            )

        for operator, n in string_properties.tilde_replacements.items():
            assert n == formatted_string.count(operator), (
                "Internal Error! Did not reverse the same number of tildes that "
                "were replaced. Please open an issue"
            )
            formatted_string = formatted_string.replace(operator, "~")

        return formatted_string


@dataclass
class JinjaTag:
    """
    A simple representation of a jinja tag.

    "verb" is one of {set, do, for, if, elif, else, test, macro, materialization, call}

    For example, "{%- set my_var=4 %}" is split into it parts:
    (opening_marker, verb, code, closing_marker) = ("{%-", "set", "my_var=4", "%}")
    """

    source_string: str
    opening_marker: str
    verb: str
    code: str
    closing_marker: str
    depth: Tuple[int, int]
    is_blackened: bool = False

    def __str__(self) -> str:
        if self.is_macro_like_def and self.is_blackened:
            self._remove_trailing_comma()

        if self.is_multiline_tag and self.is_blackened:
            return self._multiline_str()
        elif self.is_multiline_tag:
            return self.source_string
        else:
            return self._basic_str()

    @property
    def is_multiline_tag(self) -> bool:
        return "\n" in self.code

    @property
    def is_macro_like_def(self) -> bool:
        return "%" in self.opening_marker and any(
            [
                self.verb.strip() == v
                for v in ["macro", "test", "materialization", "call"]
            ]
        )

    def _multiline_str(self) -> str:
        """
        if the formatted code is on multiple lines, we want the code indented
        four spaces past the opening and closing markers. The opening marker
        will already be indented to the proper depth (because of the Line).
        """
        indent = " " * 4 * (self.depth[0] + self.depth[1])
        no_indent_lines = self._find_multiline_python_str_lines()
        code_lines = iter(self.code.splitlines(keepends=False))

        if self.verb:
            lines = [f"{self.opening_marker} {self.verb}{next(code_lines)}"]
            extra_indent = ""
        else:
            lines = [f"{self.opening_marker}"]
            extra_indent = " " * 4

        for i, code_line in enumerate(code_lines, start=1 if self.verb else 0):
            line_indent = "" if i in no_indent_lines else f"{indent}{extra_indent}"
            lines.append(f"{line_indent}{code_line}")

        if self.verb:
            lines[-1] = f"{indent}{lines[-1].lstrip()} {self.closing_marker}"
        else:
            lines.append(f"{indent}{self.closing_marker}")

        return "\n".join(lines)

    def _basic_str(self) -> str:
        return f"{self.opening_marker} {self.verb}{self.code} {self.closing_marker}"

    def _find_multiline_python_str_lines(self) -> MutableSet[int]:
        """
        Return a set line numbers that correspond with the lines
        of a triple-quoted multiline string (except for the first line
        of each string). These are lines that should never have
        their indentation adjusted
        """
        try:
            tree = ast.parse(self.code, mode="exec")
        except SyntaxError:
            # this jinja isn't quite python, so give up here.
            return set()

        line_indicies: MutableSet[int] = set()
        raw_lines = self.code.splitlines()
        for node in ast.walk(tree):
            if (
                isinstance(node, ast.Constant)
                and isinstance(node.value, str)
                and node.end_lineno is not None
                and node.end_lineno > node.lineno
                and "\n" in node.value
                and (
                    '"""' in raw_lines[node.lineno - 1]
                    or "'''" in raw_lines[node.lineno - 1]
                )
                and (
                    '"""' in raw_lines[node.end_lineno - 1]
                    or "'''" in raw_lines[node.end_lineno - 1]
                )
            ):
                line_indicies |= set(range(node.lineno, node.end_lineno))

        return line_indicies

    def _remove_trailing_comma(self) -> None:
        """
        dbt Jinja doesn't allow trailing commas in macro definitions. Mutates
        self.code so that it doesn't contain a trailing comma inside parentheses.
        """
        trailing_comma_prog = re.compile(r",\s*\)")
        source_string = self.code
        trailing_comma_match = trailing_comma_prog.search(source_string)
        if trailing_comma_match:
            idx = trailing_comma_match.span()[0]
            processed_string = f"{source_string[:idx]}{source_string[idx + 1 :]}"
            self.code = processed_string

    @classmethod
    def from_string(cls, source_string: str, depth: Tuple[int, int]) -> "JinjaTag":
        """
        Takes a jinja statement or expression as a string and returns
        a JinjaTag object (basically a tuple of its parts).
        """
        opening_marker_len = 3 if source_string[2] == "-" else 2
        opening_marker = source_string[:opening_marker_len]
        closing_marker_len = 3 if source_string[-3] == "-" else 2
        closing_marker = source_string[-closing_marker_len:]

        verb_pattern = (
            r"\s*(set|do|for|if|elif|else|test|macro|materialization|call)\s+"
        )
        verb_program = re.compile(verb_pattern, re.DOTALL | re.IGNORECASE)
        verb_match = verb_program.match(source_string[opening_marker_len:])
        if verb_match:
            verb_pos = opening_marker_len + verb_match.span(1)[0]
            verb_epos = opening_marker_len + verb_match.span(1)[1]
            verb = source_string[verb_pos:verb_epos].lower()
        else:
            verb = ""
        verb_len = verb_match.span(0)[1] - verb_match.span(0)[0] if verb_match else 0

        code_pos = opening_marker_len + verb_len
        code = source_string[code_pos:-closing_marker_len].strip()

        if verb and code:
            verb = f"{verb} "

        return JinjaTag(
            source_string, opening_marker, verb, code, closing_marker, depth
        )

    def max_code_length(self, max_length: int) -> int:
        """
        For a tag with max_length remaining characters on the line, return the
        max length that the code inside the curlies (exc. the verb) can occupy.
        """
        return (
            max_length
            - len(self.opening_marker)
            - len(self.verb)
            - len(self.closing_marker)
            - 2
        )


@dataclass
class JinjaFormatter:
    """
    Provides a simple interface, format_line, to format all jinja tags
    on a Line, using black if it is installed
    """

    mode: Mode
    code_formatter: BlackWrapper = field(default_factory=lambda: BlackWrapper())

    def __post_init__(self) -> None:
        self.use_black = (
            self.code_formatter.black is not None and not self.mode.no_jinjafmt
        )
        self.node_manager = NodeManager(self.mode.dialect.case_sensitive_names)

    def format_line(self, line: Line) -> List[Line]:
        """
        Format each jinja tag in a line, in turn. If a node was made multiline,
        split before the node (unless it is already the first node on that line)
        """
        line_length = self.mode.line_length
        if not line.formatting_disabled and line.contains_jinja:
            running_length = len(line.prefix)
            for i, node in enumerate(line.nodes):
                is_blackened = self._format_jinja_node(
                    node, max_length=line_length - running_length
                )
                if i > 0 and is_blackened and node.is_multiline_jinja:
                    # if black turned a single-line jinja expression
                    # into a multiline one, we need to split before
                    # this node (since on the first pass the splitter
                    # would have split before this node if it had been
                    # a multiline node)
                    splitter = LineSplitter(self.node_manager)
                    return list(
                        chain(
                            *[
                                self.format_line(new_line)
                                for new_line, _ in [
                                    splitter.split_at_index(line, 0, i, line.comments),
                                    splitter.split_at_index(
                                        line, i, -1, [], no_tail=True
                                    ),
                                ]
                            ]
                        )
                    )
                else:
                    running_length += len(node) - (len(node.prefix) if i == 0 else 0)
            else:
                return [line]
        else:
            return [line]

    def _format_jinja_node(self, node: Node, max_length: int) -> bool:
        """
        Format a single jinja tag. No-ops for nodes that
        are not jinja. Returns True if the node was blackened
        """
        if node.is_jinja:
            tag = JinjaTag.from_string(node.value, node.depth)

            if tag.code and self.use_black:
                tag.code, tag.is_blackened = self.code_formatter.format_string(
                    tag.code,
                    max_length=tag.max_code_length(max_length),
                )

            node.value = str(tag)

            return tag.is_blackened

        else:
            return False



import sys
from typing import Dict, Tuple

import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.jinjafmt import BlackWrapper, JinjaFormatter, JinjaTag
from sqlfmt.line import Line
from sqlfmt.mode import Mode
from sqlfmt.node_manager import NodeManager
from sqlfmt.tokens import Token, TokenType


@pytest.fixture
def node_manager(default_mode: Mode) -> NodeManager:
    return NodeManager(default_mode.dialect.case_sensitive_names)


@pytest.fixture
def jinja_formatter(default_mode: Mode) -> JinjaFormatter:
    return JinjaFormatter(default_mode)


@pytest.fixture
def no_jinjafmt_mode() -> Mode:
    return Mode(no_jinjafmt=True)


@pytest.fixture
def disabled_jinja_formatter(no_jinjafmt_mode: Mode) -> JinjaFormatter:
    return JinjaFormatter(no_jinjafmt_mode)


@pytest.fixture
def uninstall_black(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setitem(sys.modules, "black", None)


@pytest.mark.parametrize(
    "tag, result",
    [
        ("{{ expression }}", ("{{", "", "expression", "}}")),
        ("{{expression}}", ("{{", "", "expression", "}}")),
        ("{{- expression -}}", ("{{-", "", "expression", "-}}")),
        ("{{- expression }}", ("{{-", "", "expression", "}}")),
        ("{{ expression -}}", ("{{", "", "expression", "-}}")),
        ("{{expression-}}", ("{{", "", "expression", "-}}")),
        (
            "{{ my_macro(arg, kwarg=foo) }}",
            ("{{", "", "my_macro(arg, kwarg=foo)", "}}"),
        ),
        (
            "{% macro my_macro(arg, kwarg=foo) %}",
            ("{%", "macro ", "my_macro(arg, kwarg=foo)", "%}"),
        ),
        ("{% statement %}", ("{%", "", "statement", "%}")),
        ("{%- statement -%}", ("{%-", "", "statement", "-%}")),
        ("{% set a=1 %}", ("{%", "set ", "a=1", "%}")),
        ("{%set a=1%}", ("{%", "set ", "a=1", "%}")),
        ("{%-set a=1%}", ("{%-", "set ", "a=1", "%}")),
        ("{% SET a=1%}", ("{%", "set ", "a=1", "%}")),
        ("{% set my_var%}", ("{%", "set ", "my_var", "%}")),
        (
            "{% do my_list.append(something) %}",
            ("{%", "do ", "my_list.append(something)", "%}"),
        ),
    ],
)
def test_jinja_tag_from_string(tag: str, result: Tuple[str, str, str, str]) -> None:
    assert JinjaTag.from_string(tag, (0, 0)) == JinjaTag(tag, *result, (0, 0))


@pytest.mark.parametrize(
    "ttype,tag,formatted",
    [
        (TokenType.JINJA_EXPRESSION, "{{ expression }}", "{{ expression }}"),
        (TokenType.JINJA_EXPRESSION, "{{expression}}", "{{ expression }}"),
        (TokenType.JINJA_EXPRESSION, "{{- expression -}}", "{{- expression -}}"),
        (TokenType.JINJA_EXPRESSION, "{{- expression }}", "{{- expression }}"),
        (TokenType.JINJA_EXPRESSION, "{{ expression -}}", "{{ expression -}}"),
        (TokenType.JINJA_EXPRESSION, "{{expression-}}", "{{ expression -}}"),
        (
            TokenType.JINJA_EXPRESSION,
            "{{ my_macro(arg, kwarg=foo) }}",
            "{{ my_macro(arg, kwarg=foo) }}",
        ),
        (TokenType.JINJA_STATEMENT, "{% statement %}", "{% statement %}"),
        (TokenType.JINJA_STATEMENT, "{%- statement -%}", "{%- statement -%}"),
        (TokenType.JINJA_STATEMENT, "{% set a=1 %}", "{% set a = 1 %}"),
        (TokenType.JINJA_STATEMENT, "{%set a=1%}", "{% set a = 1 %}"),
        (TokenType.JINJA_STATEMENT, "{%-set a=1%}", "{%- set a = 1 %}"),
        (TokenType.JINJA_STATEMENT, "{% SET a=1%}", "{% set a = 1 %}"),
        (TokenType.JINJA_STATEMENT, "{% set my_var%}", "{% set my_var %}"),
        (
            TokenType.JINJA_STATEMENT,
            "{% do my_list.append( something ) %}",
            "{% do my_list.append(something) %}",
        ),
        # jinja allows line breaks where python doesn't; test try/except
        (
            TokenType.JINJA_STATEMENT,
            "{% set my_tuple = 1,\n    2,\n    3%}",
            "{% set my_tuple = 1, 2, 3 %}",
        ),
        # jinja allows "from" as an arg name, python doesn't
        (
            TokenType.JINJA_EXPRESSION,
            "{{dbt_utils.star(from=something)}}",
            "{{ dbt_utils.star(from=something) }}",
        ),
    ],
)
def test_format_jinja_node(
    jinja_formatter: JinjaFormatter,
    ttype: TokenType,
    tag: str,
    formatted: str,
    node_manager: NodeManager,
) -> None:
    t = Token(type=ttype, prefix="", token=tag, spos=0, epos=len(tag))
    n = node_manager.create_node(t, previous_node=None)
    jinja_formatter._format_jinja_node(n, 88)
    assert n.value == formatted


@pytest.mark.parametrize(
    "source_string",
    [
        "{% set a=1 %}",
        "{% set a    = 1 %}",
        '{{\n    dbt_utils.date_spine(\n        datepart="day",\n    )\n}}',
    ],
)
def test_no_format_jinja_node(
    disabled_jinja_formatter: JinjaFormatter,
    source_string: str,
    node_manager: NodeManager,
) -> None:
    t = Token(
        type=TokenType.JINJA_STATEMENT,
        prefix="",
        token=source_string,
        spos=0,
        epos=len(source_string),
    )
    n = node_manager.create_node(t, previous_node=None)
    disabled_jinja_formatter._format_jinja_node(n, 88)
    assert n.value == source_string


@pytest.mark.parametrize(
    "source_string",
    [
        "{% set a=1 %}",
        "{% set a    = 1 %}",
        '{{\n    dbt_utils.date_spine(\n        datepart="day",\n    )\n}}',
    ],
)
def test_format_jinja_node_no_black(
    uninstall_black: None,
    jinja_formatter: JinjaFormatter,
    source_string: str,
    node_manager: NodeManager,
) -> None:
    t = Token(
        type=TokenType.JINJA_STATEMENT,
        prefix="",
        token=source_string,
        spos=0,
        epos=len(source_string),
    )
    n = node_manager.create_node(t, previous_node=None)
    jinja_formatter._format_jinja_node(n, 88)
    assert n.value == source_string


def test_format_line(
    jinja_formatter: JinjaFormatter, node_manager: NodeManager
) -> None:
    t = Token(
        type=TokenType.JINJA_EXPRESSION,
        prefix="",
        token="{{expression}}",
        spos=0,
        epos=14,
    )
    n = node_manager.create_node(t, previous_node=None)
    line = Line.from_nodes(previous_node=None, nodes=[n], comments=[])
    node_manager.append_newline(line)

    assert n.value == "{{expression}}"
    _ = list(map(jinja_formatter.format_line, [line]))
    assert n.value == "{{ expression }}"


def test_format_line_single_to_multi(
    default_analyzer: Analyzer, jinja_formatter: JinjaFormatter
) -> None:
    source_string = (
        "from {{ source('fooooooooooooooooooooooooooooooooooooooooooooooooooo', "
        "'barrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrrr') }}"
    )
    q = default_analyzer.parse_query(source_string=source_string)
    assert len(q.lines) == 1
    line = q.lines[0]
    j_node = line.nodes[1]
    assert j_node.is_jinja
    assert not j_node.is_multiline_jinja

    new_lines = jinja_formatter.format_line(line)
    assert len(new_lines) == 2  # make sure the linesplitter did its thing
    assert j_node.is_multiline_jinja  # jinjafmt mutated this node to be multiline


@pytest.mark.parametrize(
    "source_string,expected_string",
    [
        ("1 +      1", "1 + 1"),
        ("'a' ~      'b'", '"a" ~ "b"'),
        ("dbt_utils.star(\nfrom=one\n)", "dbt_utils.star(from=one)"),
        ("return(\nfoo\n)", "return(foo)"),
    ],
)
def test_black_wrapper_format_string(
    source_string: str, expected_string: str, jinja_formatter: JinjaFormatter
) -> None:
    result = jinja_formatter.code_formatter.format_string(
        source_string=source_string, max_length=88
    )
    assert result == (expected_string, True)


def test_black_wrapper_format_string_no_black(
    uninstall_black: None, default_mode: Mode
) -> None:
    source_string = "1 +    1"
    jinja_formatter = JinjaFormatter(default_mode)
    result = jinja_formatter.code_formatter.format_string(
        source_string=source_string, max_length=88
    )
    assert result == (source_string, False)


@pytest.mark.parametrize("source_string", [":::", ":::\n:::", "1,\n2return(return_())"])
def test_black_wrapper_format_string_invalid_input(
    source_string: str,
    jinja_formatter: JinjaFormatter,
) -> None:
    """
    Our prepocessor will abort the keyword substitution because the sentinel return_
    already exists in the source.
    """
    result = jinja_formatter.code_formatter.format_string(
        source_string=source_string, max_length=88
    )
    assert result == (source_string, False)


@pytest.mark.parametrize(
    "tag,expected",
    [
        (
            JinjaTag("{{ any code! }}", "{{", "", "any code!", "}}", (0, 0)),
            88 - 2 - 2 - 2,
        ),
        (JinjaTag("{%- set _ %}", "{%-", "set", "_", "%}", (0, 0)), 88 - 3 - 3 - 2 - 2),
    ],
)
def test_max_code_length(tag: JinjaTag, expected: int) -> None:
    result = tag.max_code_length(88)
    assert result == expected


@pytest.mark.parametrize(
    "source_string,expected_res",
    [
        ("my_macro(one, two, three='my_kwarg')", ("", (False, {}, {}))),
        ("my_macro(\n    one, two, three='my_kwarg'\n)", ("", (True, {}, {}))),
        (
            "my_macro(\n    one,\n    two,\n    three='my_kwarg',\n)",
            ("", (True, {}, {})),
        ),
        ("my_list = [\n1, 2, 3,\n]", ("", (True, {}, {}))),
        (
            "return(import(except=1))",
            (
                "return_(import_(except_=1))",
                (False, {r"return_\(": 1, r"import_\(": 1, r"except_\s*=": 1}, {}),
            ),
        ),
        ("'a' ~     'b'", ("'a' +     'b'", (False, {}, {"+": 1}))),
    ],
)
def test_preprocess_string_properties(
    source_string: str, expected_res: Tuple[str, BlackWrapper.StringProperties]
) -> None:
    res_string, res_properties = BlackWrapper._preprocess_string(source_string)
    assert res_string == expected_res[0] if expected_res[0] else source_string
    assert res_properties == expected_res[1]


@pytest.mark.parametrize(
    "source_string,expected_count,is_def",
    [
        ("{% macro my_macro(foo, bar, baz=qux) %}", 2, True),
        ("{% test my_test(foo, bar, baz=qux) %}", 2, True),
        ("{% macro my_macro(foo, bar, baz=qux,) %}", 2, True),
        ("{% test my_test(foo, bar, baz=qux,) %}", 2, True),
        ("{% macro my_macro(\n    foo,\n    bar, \n    baz=qux,\n) %}", 2, True),
        ("{% test my_test(\n    foo,\n    bar, \n    baz=qux,\n) %}", 2, True),
        ("{% materialization my_mat, default %}", 1, True),
        ("{% call statement('foo', fetch_results=True, auto_begin=True) %}", 2, True),
        # do not remove from macro calls, only defs
        ("{{ my_macro(\n    foo,\n    bar, \n    baz=qux,\n) }}", 3, False),
    ],
)
def test_jinja_tag_remove_trailing_comma(
    source_string: str, expected_count: int, is_def: bool
) -> None:
    tag = JinjaTag.from_string(source_string=source_string, depth=(0, 0))
    assert tag.is_macro_like_def == is_def
    tag.is_blackened = True
    result = str(tag)
    assert result.count(",") == expected_count


@pytest.mark.parametrize(
    "source_string,expected_string,expected_kw_replacements",
    [
        (
            "return(adapter.dispatch('my_macro', 'my_package')(arg1, arg2))",
            "return_(adapter.dispatch('my_macro', 'my_package')(arg1, arg2))",
            {r"return_\(": 1},
        ),
        (
            "dbt_utils.star(from=ref('asdf'), except=fields_to_exclude)",
            "dbt_utils.star(from_=ref('asdf'), except_=fields_to_exclude)",
            {r"from_\s*=": 1, r"except_\s*=": 1},
        ),
        (
            "True = foo",
            "True_= foo",
            {r"True_\s*=": 1},
        ),
        (
            "abc = True",
            "abc = True",
            {},
        ),
        (
            "if foo == 'bar'",
            "if foo == 'bar'",
            {},
        ),
        (
            "from_ = 'already there'",
            "from_ = 'already there'",
            {},
        ),
        (
            "from_ = from()",
            "from_ = from_()",
            {r"from_\(": 1},
        ),
        (
            "return(from_())",
            "return_(from_())",
            {r"return_\(": 1},
        ),
    ],
)
def test_replace_reserved_words(
    source_string: str, expected_string: str, expected_kw_replacements: Dict[str, int]
) -> None:
    processed_string, actual_kw_replacements = BlackWrapper._replace_reserved_words(
        source_string=source_string
    )
    assert processed_string == expected_string
    assert actual_kw_replacements == expected_kw_replacements


@pytest.mark.parametrize(
    "source_string",
    [
        "from_ = (from == 1)",
        "return(return_())",
    ],
)
def test_replace_reserved_words_preexisting_sentinels(source_string: str) -> None:
    with pytest.raises(ValueError):
        _, _ = BlackWrapper._replace_reserved_words(source_string=source_string)


@pytest.mark.parametrize(
    "source_string,expected_string,expected_tilde_replacements",
    [
        ('"a" ~ "b"', '"a" + "b"', {"+": 1}),
        ('"a" ~ "b" + 1', '"a" - "b" + 1', {"-": 1}),
        ('"a" ~ "b" - 1', '"a" + "b" - 1', {"+": 1}),
        ("foo", "foo", {}),
        ("foo + bar", "foo + bar", {}),
        ("~+-*/", "~+-*/", {}),
        ("foo ~ bar + baz - qux * foo / bar", "foo ~ bar + baz - qux * foo / bar", {}),
    ],
)
def test_replace_tildes(
    source_string: str,
    expected_string: str,
    expected_tilde_replacements: Dict[str, int],
) -> None:
    processed_string, actual_tilde_replacements = BlackWrapper._replace_tildes(
        source_string=source_string
    )
    assert processed_string == expected_string
    assert actual_tilde_replacements == expected_tilde_replacements


@pytest.mark.parametrize(
    "formatted_string,keyword_replacements,tilde_replacements,expected_string",
    [
        (
            "from_ = 'already there'",
            {},
            {},
            "from_ = 'already there'",
        ),
        (
            "from_ = 'replaced a keyword'",
            {r"from_\s*=": 1},
            {},
            "from = 'replaced a keyword'",
        ),
        (
            "dbt_utils.star(from_=ref('asdf'), except_=fields_to_exclude)",
            {r"from_\s*=": 1, r"except_\s*=": 1},
            {},
            "dbt_utils.star(from=ref('asdf'), except=fields_to_exclude)",
        ),
        ("foo + bar", {}, {"+": 1}, "foo ~ bar"),
        ("foo + bar / baz", {}, {"/": 1}, "foo + bar ~ baz"),
        ("foo + bar", {}, {}, "foo + bar"),
    ],
)
def test_postprocess_string(
    formatted_string: str,
    keyword_replacements: Dict[str, int],
    tilde_replacements: Dict[str, int],
    expected_string: str,
) -> None:
    processed_string = BlackWrapper._postprocess_string(
        formatted_string,
        string_properties=BlackWrapper.StringProperties(
            has_newlines=False,
            keyword_replacements=keyword_replacements,
            tilde_replacements=tilde_replacements,
        ),
    )
    assert processed_string == expected_string


@pytest.mark.parametrize(
    "source_string",
    BlackWrapper.PY_RESERVED_WORDS
    + [
        "dbt_utils.star(from=ref('asdf'), except=fields_to_exclude)",
        "return(adapter.dispatch('my_macro', 'my_package')(arg1, arg2))",
        "True = foo",
        "abc = True",
        "something if something_else",
        "foo ~ bar",
    ],
)
def test_preprocess_and_postprocess_are_inverse_ops(source_string: str) -> None:
    """
    Preprocess and Postprocess should be perfectly inverse operations, except
    for some whitespace (we don't need to be too precious about whitespace because
    in the app Black will run between pre and post processing)
    """
    assert BlackWrapper._postprocess_string(
        *BlackWrapper._preprocess_string(source_string)
    ).replace(" ", "") == source_string.replace(" ", "")


@pytest.mark.parametrize(
    "source_string",
    [
        """{{\n    config(\n        foo="bar",\n    )\n}}""",
        '''{{\n    config(\n        foo="""\n\nbar\n\n""",\n    )\n}}''',
    ],
)
def test_multiline_str(source_string: str) -> None:
    tag = JinjaTag.from_string(source_string=source_string, depth=(0, 0))
    tag.code, tag.is_blackened = BlackWrapper().format_string(
        source_string=tag.code, max_length=88
    )
    assert tag.is_blackened
    assert str(tag) == source_string



from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from sqlfmt.comment import Comment
from sqlfmt.node import Node, get_previous_token
from sqlfmt.tokens import Token, TokenType


@dataclass
class Line:
    """
    A Line is a collection of Nodes and Comments that should be printed together, on a
    single line.
    """

    previous_node: Optional[Node]  # last node of prior line, if any
    nodes: List[Node] = field(default_factory=list)
    comments: List[Comment] = field(default_factory=list)
    formatting_disabled: List[Token] = field(default_factory=list)

    def __str__(self) -> str:
        """
        A Line is printed in one of three ways:
        1. Blank lines are just bare newlines, with no other whitespace
        2. Lines where formatting is disabled must use the original lexed token,
           and print exactly what we lexed
        3. Concatenate all Nodes and prepend the correct amount of whitespace
           for indentation

        Does not include any Comments; for those, use the render_with_comments
        method
        """
        if self.is_blank_line:
            return "\n"
        elif self.formatting_disabled:
            return "".join([f"{t.prefix}{t.token}" for t in self.tokens])
        else:
            return self.prefix + "".join([str(node) for node in self.nodes]).lstrip(" ")

    def __len__(self) -> int:
        try:
            return max([len(s) for s in str(self).splitlines()])
        except ValueError:
            return 0

    @property
    def open_brackets(self) -> List[Node]:
        """
        The brackets open at the start of this Line
        """
        if self.nodes:
            return self.nodes[0].open_brackets
        elif self.previous_node is not None:
            return self.previous_node.open_brackets
        else:
            return []

    @property
    def open_jinja_blocks(self) -> List[Node]:
        """
        The jinja blocks open at the start of this Line
        """
        if self.nodes:
            return self.nodes[0].open_jinja_blocks
        elif self.previous_node is not None:
            return self.previous_node.open_jinja_blocks
        else:
            return []

    @property
    def depth(self) -> Tuple[int, int]:
        """
        The depth of the start of this line
        """
        if self.nodes:
            return (len(self.open_brackets), len(self.open_jinja_blocks))
        else:
            return (0, 0)

    @property
    def prefix(self) -> str:
        """
        Returns the whitespace to be printed at the start of this Line for
        proper indentation.
        """
        INDENT = " " * 4
        prefix = INDENT * (self.depth[0] + self.depth[1])
        return prefix

    def render_with_comments(self, max_length: int) -> str:
        """
        Returns a string that represents the properly-formatted Line,
        including associated comments
        """
        content = str(self).rstrip()
        rendered_lines: List[str] = []
        inline_comments: List[str] = []
        for comment in self.comments:
            if comment.is_multiline or comment.is_standalone:
                rendered_lines.append(
                    comment.render_standalone(max_length=max_length, prefix=self.prefix)
                )
            else:
                inline_comments.append(comment.render_inline())

        if inline_comments:
            rendered_lines.append(f"{content}{''.join(inline_comments)}\n")
        # if content is empty AND we've already rendered a standalone or multiline
        # comment, skip rendering the line itself, since that just adds a duplicate
        # newline.
        elif content or not self.comments:
            rendered_lines.append(f"{self}")

        return "".join(rendered_lines)

    @classmethod
    def from_nodes(
        cls,
        previous_node: Optional[Node],
        nodes: List[Node],
        comments: List[Comment],
    ) -> "Line":
        """
        Creates and returns a new line from a list of Nodes. Useful for line
        splitting and merging
        """
        if nodes:
            line = Line(
                previous_node=previous_node,
                nodes=nodes,
                comments=comments,
                formatting_disabled=nodes[0].formatting_disabled
                + nodes[-1].formatting_disabled,
            )
        else:
            line = Line(
                previous_node=previous_node,
                nodes=nodes,
                comments=comments,
                formatting_disabled=(
                    previous_node.formatting_disabled
                    if previous_node is not None
                    else []
                ),
            )

        return line

    @property
    def tokens(self) -> List[Token]:
        tokens = []
        for node in self.nodes:
            tokens.append(node.token)
        return tokens

    @property
    def previous_token_is_comma(self) -> bool:
        token, _ = get_previous_token(self.previous_node)
        if token is None:
            return False
        else:
            return token.type is TokenType.COMMA

    @property
    def is_blank_line(self) -> bool:
        if (
            len(self.nodes) == 1
            and self.nodes[0].is_newline
            and len(self.comments) == 0
        ):
            return True
        else:
            return False

    @property
    def starts_with_unterm_keyword(self) -> bool:
        try:
            return self.nodes[0].is_unterm_keyword
        except IndexError:
            return False

    @property
    def starts_with_operator(self) -> bool:
        try:
            return self.nodes[0].is_operator
        except IndexError:
            return False

    @property
    def starts_with_comma(self) -> bool:
        try:
            return self.nodes[0].is_comma
        except IndexError:
            return False

    @property
    def starts_with_jinja_statement(self) -> bool:
        try:
            return self.nodes[0].is_jinja_statement
        except IndexError:
            return False

    @property
    def starts_with_bracket_operator(self) -> bool:
        try:
            return self.nodes[0].is_bracket_operator
        except IndexError:
            return False

    @property
    def contains_unterm_keyword(self) -> bool:
        return any([n.is_unterm_keyword for n in self.nodes])

    @property
    def contains_operator(self) -> bool:
        return any([n.is_operator for n in self.nodes])

    @property
    def contains_jinja(self) -> bool:
        return any([n.is_jinja for n in self.nodes])

    @property
    def is_standalone_jinja_statement(self) -> bool:
        return self._is_standalone_if(self.starts_with_jinja_statement)

    @property
    def is_standalone_operator(self) -> bool:
        return self._is_standalone_if(
            self.starts_with_operator and not self.starts_with_bracket_operator
        )

    @property
    def is_standalone_comma(self) -> bool:
        return self._is_standalone_if(self.starts_with_comma)

    def _is_standalone_if(self, starts_with_type: bool) -> bool:
        if len(self.nodes) == 1 and starts_with_type:
            return True
        if len(self.nodes) == 2 and starts_with_type and self.nodes[1].is_newline:
            return True
        else:
            return False

    def is_too_long(self, max_length: int) -> bool:
        """
        Returns true if the rendered length of the line is strictly greater
        than max_length, and if the line isn't a standalone long
        multiline node
        """
        if len(self) > max_length:
            return True
        else:
            return False

    @property
    def closes_bracket_from_previous_line(self) -> bool:
        """
        Returns true for a line with an explicit bracket like ")" or "]"
        that matches a bracket on a preceding line. False for unterminated
        keywords or any lines with matched brackets
        """
        if (
            self.previous_node is not None
            and self.previous_node.open_brackets
            and self.nodes
        ):
            explicit_brackets = [
                b for b in self.previous_node.open_brackets if b.is_opening_bracket
            ]
            if (
                explicit_brackets
                and explicit_brackets[-1] not in self.nodes[-1].open_brackets
            ):
                return True
        return False

    @property
    def previous_line_has_open_jinja_blocks_not_keywords(self) -> bool:
        """
        Returns true if the previous line is inside a jinja block, but not
        after a jinja block keyword, like {% else %}/{% elif %}
        """
        if (
            self.previous_node is not None
            and self.previous_node.open_jinja_blocks
            and not self.previous_node.open_jinja_blocks[-1].is_jinja_block_keyword
        ):
            return True
        else:
            return False

    @property
    def closes_jinja_block_from_previous_line(self) -> bool:
        """
        Returns true for a line that contains {% endif %}, {% endfor %}, etc.
        where the matching {% if %}, etc. is on a previous line.
        """
        if (
            self.nodes
            and self.previous_node is not None
            and self.previous_node.open_jinja_blocks
            and (
                self.previous_node.open_jinja_blocks[-1]
                not in self.nodes[-1].open_jinja_blocks
            )
            and (
                self.nodes[-1].open_jinja_blocks == []
                or not self.nodes[-1].open_jinja_blocks[-1].is_jinja_block_keyword
            )
        ):
            return True
        return False

    @property
    def closes_simple_jinja_block_from_previous_line(self) -> bool:
        """
        Returns true for a line that contains {% endif %}, {% endfor %}, etc.
        where the matching {% if %}, etc. is on a previous line. Returns False
        for {% else %}/{% elif %} or an {% endif %} that follows an {% else %}/
        {% endif %}
        """
        if (
            self.previous_line_has_open_jinja_blocks_not_keywords
            and self.closes_jinja_block_from_previous_line
        ):
            return True
        return False

    @property
    def opens_new_bracket(self) -> bool:
        """
        Returns True iff the Nodes in this Line open a new explicit bracket and do
        not also close that bracket
        """
        if not self.nodes:
            return False
        elif not self.nodes[-1].open_brackets:
            return False
        else:
            b = self.nodes[-1].open_brackets[-1]
            if b.is_opening_bracket and b not in self.open_brackets:
                return True
            else:
                return False

    def starts_new_segment(self, prev_segment_depth: Tuple[int, int]) -> bool:
        if self.depth <= prev_segment_depth or self.depth[1] < prev_segment_depth[1]:
            # if this line starts with a closing bracket,
            # we want to include that closing bracket
            # in the same segment as the first line.
            if (
                self.closes_bracket_from_previous_line
                or self.closes_simple_jinja_block_from_previous_line
                or self.is_blank_line
            ) and self.depth == prev_segment_depth:
                return False
            else:
                return True
        return False



import re
import sys
from enum import Enum, auto
from typing import NamedTuple

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from backports.cached_property import cached_property


class TokenType(Enum):
    """
    Tokens are basic elements of SQL syntax; we only care about a small number of
    token types, which correspond to query structure and therefore formatting
    """

    FMT_OFF = auto()
    FMT_ON = auto()
    DATA = auto()
    JINJA_STATEMENT = auto()  # {% ... %}
    JINJA_EXPRESSION = auto()  # {{ ... }}
    JINJA_BLOCK_START = auto()  # {% if ... %}
    JINJA_BLOCK_END = auto()  # {% endif %}
    JINJA_BLOCK_KEYWORD = auto()  # {% else %}
    QUOTED_NAME = auto()
    COMMENT = auto()
    COMMENT_START = auto()
    COMMENT_END = auto()
    SEMICOLON = auto()
    STATEMENT_START = auto()
    STATEMENT_END = auto()
    STAR = auto()
    NUMBER = auto()
    BRACKET_OPEN = auto()
    BRACKET_CLOSE = auto()
    DOUBLE_COLON = auto()
    COLON = auto()
    OPERATOR = auto()
    WORD_OPERATOR = auto()
    ON = auto()
    BOOLEAN_OPERATOR = auto()
    COMMA = auto()
    DOT = auto()
    NEWLINE = auto()
    UNTERM_KEYWORD = auto()  # Unterminated keyword
    SET_OPERATOR = auto()
    NAME = auto()

    @cached_property
    def is_jinja_statement(self) -> bool:
        return self in (
            TokenType.JINJA_STATEMENT,
            TokenType.JINJA_BLOCK_START,
            TokenType.JINJA_BLOCK_KEYWORD,
            TokenType.JINJA_BLOCK_END,
        )

    @cached_property
    def is_jinja(self) -> bool:
        return self.is_jinja_statement or self is TokenType.JINJA_EXPRESSION

    @cached_property
    def is_opening_bracket(self) -> bool:
        return self in [
            TokenType.BRACKET_OPEN,
            TokenType.STATEMENT_START,
        ]

    @cached_property
    def divides_queries(self) -> bool:
        return self in [
            TokenType.SEMICOLON,
            TokenType.SET_OPERATOR,
        ]

    @cached_property
    def does_not_set_prev_sql_context(self) -> bool:
        return self.is_jinja_statement or self is TokenType.NEWLINE

    @cached_property
    def is_always_operator(self) -> bool:
        return self in [
            TokenType.OPERATOR,
            TokenType.WORD_OPERATOR,
            TokenType.ON,
            TokenType.BOOLEAN_OPERATOR,
            TokenType.DOUBLE_COLON,
            TokenType.SEMICOLON,
        ]

    @cached_property
    def is_never_preceded_by_space(self) -> bool:
        return self in [
            TokenType.BRACKET_CLOSE,
            TokenType.COLON,
            TokenType.DOUBLE_COLON,
            TokenType.COMMA,
            TokenType.DOT,
            TokenType.NEWLINE,
        ]

    @cached_property
    def is_preceded_by_space_except_after_open_bracket(self) -> bool:
        return self in [
            TokenType.UNTERM_KEYWORD,
            TokenType.STATEMENT_START,
            TokenType.STATEMENT_END,
            TokenType.WORD_OPERATOR,
            TokenType.BOOLEAN_OPERATOR,
            TokenType.ON,
        ]

    @cached_property
    def is_possible_name(self) -> bool:
        return self in [
            TokenType.QUOTED_NAME,
            TokenType.NAME,
            TokenType.STAR,
            TokenType.JINJA_EXPRESSION,
        ]

    @cached_property
    def is_always_lowercased(self) -> bool:
        return self in [
            TokenType.UNTERM_KEYWORD,
            TokenType.BRACKET_OPEN,
            TokenType.STATEMENT_START,
            TokenType.STATEMENT_END,
            TokenType.WORD_OPERATOR,
            TokenType.ON,
            TokenType.BOOLEAN_OPERATOR,
            TokenType.SET_OPERATOR,
            TokenType.NUMBER,
        ]

    @cached_property
    def is_equivalent_in_output(self) -> bool:
        return self not in [
            TokenType.NEWLINE,
            TokenType.COMMENT,
        ]


class Token(NamedTuple):
    """
    Representation of a syntactic element. Tokens always reference their position
    in the source query, and contain the full text of the line from the source
    query. Tokens are immutable, but some operations replace Tokens in order to
    update their type, prefix, or token in the formatted result.
    """

    type: TokenType
    prefix: str
    token: str
    spos: int
    epos: int

    def __str__(self) -> str:
        return f"Token(type={self.type}, token={self.token}, spos={self.spos})"

    def __repr__(self) -> str:
        return (
            f"Token(type={self.type}, "
            f"prefix={repr(self.prefix)}, token={repr(self.token)}, "
            f"spos={self.spos}, epos={self.epos}"
            f")"
        )

    @classmethod
    def from_match(
        cls,
        source_string: str,
        match: re.Match,
        token_type: TokenType,
    ) -> "Token":
        """
        Constructs a Token based on a regex match in a source string, and a type.
        """
        pos, _ = match.span(0)
        spos, epos = match.span(1)
        prefix = source_string[pos:spos]
        token_text = source_string[spos:epos]
        return Token(token_type, prefix, token_text, pos, epos)



import pickle
from pathlib import Path
from typing import Dict, Generator, List, Tuple

import pytest

from sqlfmt.cache import (
    Cache,
    check_cache,
    clear_cache,
    get_cache_file,
    load_cache,
    write_cache,
)
from sqlfmt.exception import SqlfmtError
from sqlfmt.mode import Mode
from sqlfmt.report import SqlFormatResult
from tests.util import BASE_DIR


@pytest.fixture(autouse=True)
def auto_clear_cache() -> Generator[None, None, None]:
    clear_cache()
    yield
    clear_cache()


@pytest.fixture
def sample_paths() -> Dict[str, Path]:
    paths = {
        "001": BASE_DIR / "preformatted" / "001_select_1.sql",
        "002": BASE_DIR / "preformatted" / "002_select_from_where.sql",
        "003": BASE_DIR / "preformatted" / "003_literals.sql",
        "004": BASE_DIR / "preformatted" / "004_with_select.sql",
        "005": BASE_DIR / "preformatted" / "005_fmt_off.sql",
        "900": BASE_DIR / "errors" / "900_bad_token.sql",
    }
    return paths


@pytest.fixture
def sample_stat() -> Tuple[float, int]:
    return (1000000000.100000, 1)


@pytest.fixture
def small_cache(sample_paths: Dict[str, Path], sample_stat: Tuple[float, int]) -> Cache:
    cache = {v: sample_stat for v in sample_paths.values() if "errors" not in str(v)}
    return cache


@pytest.fixture
def results_for_caching(sample_paths: Dict[str, Path]) -> List[SqlFormatResult]:
    results = [
        SqlFormatResult(
            sample_paths["001"],
            "select 1\n",
            "select 1\n",
            encoding="utf-8",
            utf_bom="",
        ),
        SqlFormatResult(
            sample_paths["002"],
            "select 1\n",
            "",
            encoding="utf-8",
            utf_bom="",
            from_cache=True,
        ),
        SqlFormatResult(
            sample_paths["003"],
            "select 'abc'\n",
            "select\n    'abc'\n",
            encoding="utf-8",
            utf_bom="",
        ),
        SqlFormatResult(
            sample_paths["900"],
            "!\n",
            "",
            encoding="utf-8",
            utf_bom="",
            exception=SqlfmtError("oops"),
        ),
    ]
    return results


def test_get_cache_file() -> None:
    cache_file = get_cache_file()
    assert cache_file
    assert isinstance(cache_file, Path)


def test_write_cache(
    small_cache: Cache,
    results_for_caching: List[SqlFormatResult],
    default_mode: Mode,
    sample_paths: Dict[str, Path],
    sample_stat: Tuple[float, int],
) -> None:
    cache_file = get_cache_file()
    assert not cache_file.exists()
    write_cache(cache=small_cache, results=results_for_caching, mode=default_mode)
    assert cache_file.exists()
    with open(cache_file, "rb") as f:
        written_cache = pickle.load(f)
    assert isinstance(written_cache, dict)
    assert small_cache.keys() == written_cache.keys()
    assert written_cache[sample_paths["001"]] != sample_stat, (
        "Should write new stat to cache for unchanged files"
    )
    assert written_cache[sample_paths["002"]] == sample_stat, (
        "Should not write new stat to cache for results from cache"
    )
    assert written_cache[sample_paths["003"]] != sample_stat, (
        "Should write new stat to cache for changed files in default mode"
    )
    assert sample_paths["900"] not in written_cache, "Should not write errors to cache"


def test_load_cache(
    small_cache: Cache,
    results_for_caching: List[SqlFormatResult],
    default_mode: Mode,
) -> None:
    empty_cache = load_cache()
    assert empty_cache == {}
    write_cache(cache=small_cache, results=results_for_caching, mode=default_mode)
    populated_cache = load_cache()
    assert isinstance(populated_cache, dict)
    assert populated_cache.keys() == small_cache.keys()
    assert populated_cache != small_cache


def test_check_cache(
    small_cache: Cache,
    results_for_caching: List[SqlFormatResult],
    default_mode: Mode,
    sample_paths: Dict[str, Path],
) -> None:
    assert all([not check_cache(small_cache, p) for p in sample_paths.values()])
    write_cache(cache=small_cache, results=results_for_caching, mode=default_mode)
    new_cache = load_cache()
    expected_cache_hits = [
        True,
        False,
        True,
        False,
        False,
        False,
    ]
    actual_cache_hits = [check_cache(new_cache, p) for p in sample_paths.values()]
    assert actual_cache_hits == expected_cache_hits


def test_clear_cache(
    small_cache: Cache,
    results_for_caching: List[SqlFormatResult],
    default_mode: Mode,
) -> None:
    cache_path = get_cache_file()
    write_cache(cache=small_cache, results=results_for_caching, mode=default_mode)
    assert cache_path.exists()

    clear_cache()
    assert not cache_path.exists()



import pytest

from sqlfmt.dialect import ClickHouse, Polyglot
from sqlfmt.exception import SqlfmtConfigError
from sqlfmt.mode import Mode


@pytest.mark.parametrize(
    "no_color,force_color,env_no_color,result",
    [
        (True, False, False, False),
        (False, False, True, False),
        (True, False, True, False),
        (True, True, True, True),
        (True, True, False, True),
        (False, True, True, True),
        (False, False, False, True),
    ],
)
def test_color_mode(
    no_color: bool,
    force_color: bool,
    env_no_color: bool,
    result: bool,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    if env_no_color:
        monkeypatch.setenv("NO_COLOR", "1")
    else:
        monkeypatch.delenv("NO_COLOR", raising=False)

    mode = Mode(no_color=no_color, force_color=force_color)

    assert mode.color is result


def test_dialect() -> None:
    mode = Mode()
    assert isinstance(mode.dialect, Polyglot)

    clickhouse_mode = Mode(dialect_name="clickhouse")
    # clickhouse is a subclass of Polyglot
    assert isinstance(clickhouse_mode.dialect, Polyglot)
    assert isinstance(clickhouse_mode.dialect, ClickHouse)


def test_dialect_raises() -> None:
    # clickhouse is a subclass of Polyglot
    with pytest.raises(SqlfmtConfigError):
        _ = Mode(dialect_name="foo")



import shutil
import timeit
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional

import click
from git.repo import Repo
from platformdirs import user_cache_dir

from sqlfmt.api import get_matching_paths, initialize_progress_bar, run
from sqlfmt.cache import get_cache_file
from sqlfmt.mode import Mode


@dataclass
class SQLProject:
    name: str
    git_url: str
    git_ref: str
    expected_changed: int
    expected_unchanged: int
    expected_errored: int
    sub_directory: Optional[Path] = None


def get_projects() -> List[SQLProject]:
    projects = [
        SQLProject(
            name="gitlab",
            git_url="https://github.com/tconbeer/gitlab-analytics-sqlfmt.git",
            git_ref="f859b79",  # sqlfmt bc43911
            expected_changed=1,
            expected_unchanged=2416,
            expected_errored=0,
            sub_directory=Path("transform/snowflake-dbt/"),
        ),
        SQLProject(
            name="rittman",
            git_url="https://github.com/tconbeer/rittman_ra_data_warehouse.git",
            git_ref="5fe702b",  # sqlfmt bc43911
            expected_changed=0,
            expected_unchanged=307,
            expected_errored=4,  # true mismatching brackets
            sub_directory=Path(""),
        ),
        SQLProject(
            name="http_archive",
            git_url="https://github.com/tconbeer/http_archive_almanac.git",
            git_ref="10f3238",  # sqlfmt e416385
            expected_changed=0,
            expected_unchanged=1729,
            expected_errored=0,
            sub_directory=Path("sql"),
        ),
        SQLProject(
            name="aqi",
            git_url="https://github.com/tconbeer/aqi_livibility_analysis.git",
            git_ref="f313f5b",  # sqlfmt 54b8edd
            expected_changed=0,
            expected_unchanged=7,
            expected_errored=0,
            sub_directory=Path("src/aqi_dbt/models"),
        ),
        SQLProject(
            name="jaffle_shop",
            git_url="https://github.com/tconbeer/jaffle_shop.git",
            git_ref="5e63c7c",  # sqlfmt 54b8edd
            expected_changed=0,
            expected_unchanged=5,
            expected_errored=0,
            sub_directory=Path(""),
        ),
        SQLProject(
            name="dbt_utils",
            git_url="https://github.com/tconbeer/dbt-utils.git",
            git_ref="4ce4e36",  # sqlfmt 0622c4e
            expected_changed=0,
            expected_unchanged=131,
            expected_errored=0,
            sub_directory=Path(""),
        ),
    ]
    return projects


@click.command()
@click.argument("project_names", nargs=-1)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help=("Does not print sqlfmt report to stderr"),
)
@click.option(
    "-k",
    "--reset-cache",
    is_flag=True,
    help=("Always does a fresh clone from project's repo."),
)
@click.option(
    "--single-process",
    is_flag=True,
    help=("Run sqlfmt in a single process. Useful for profiling"),
)
@click.pass_context
def sqlfmt_primer(
    ctx: click.Context,
    quiet: bool,
    reset_cache: bool,
    single_process: bool,
    project_names: List[str],
) -> None:
    """
    Run sqlfmt against one or many projects.

    With no arguments, runs against all projects. Otherwise, valide PROJECT NAMES
    are currently: "aqi", "gitlab", "rittman", and "http_archive"
    """

    all_projects = get_projects()
    all_names = [project.name for project in all_projects]
    if project_names:
        for proj_name in project_names:
            if proj_name not in all_names:
                click.echo(f"FATAL: {proj_name} not in known projects: {all_names}")
                ctx.exit(2)

    # ensure we have sqlfmt and git installed
    git = shutil.which("git")
    assert git, "git must be installed before running primer"

    if not project_names:
        projects = all_projects
    else:
        projects = [
            project for project in all_projects if project.name in project_names
        ]

    mode = Mode(quiet=True, check=True, single_process=single_process)
    exit_code = 0
    clear_sqlfmt_cache()

    with TemporaryDirectory() as working_dir:
        for project in projects:
            target_dir = get_project_source_tree(project, reset_cache, working_dir)

            click.echo(f"Running sqlfmt on {project.name}", err=True)

            files = get_matching_paths([target_dir], mode=mode)
            progress_bar, progress_callback = initialize_progress_bar(
                total=len(files), mode=mode
            )

            start_time = timeit.default_timer()
            report = run(files=files, mode=mode, callback=progress_callback)
            end_time = timeit.default_timer()

            progress_bar.close()

            number_formatted = (
                report.number_changed + report.number_unchanged + report.number_errored
            )
            time_elapsed = end_time - start_time
            click.echo(
                (
                    f"Run completed in {time_elapsed:.2f} seconds "
                    f"({number_formatted / time_elapsed:.1f} files/sec)"
                ),
                err=True,
            )

            if not quiet:
                report.display_report()

            if report.number_changed != project.expected_changed:
                _warn(project.name)
                _warn(
                    f"Changed:: Expected: "
                    f"{project.expected_changed}, Actual: {report.number_changed}"
                )
            if report.number_unchanged != project.expected_unchanged:
                _warn(
                    f"Unchanged:: Expected: "
                    f"{project.expected_unchanged}, "
                    f"Actual: {report.number_unchanged}"
                )
            if report.number_errored != project.expected_errored:
                _warn(
                    f"Errored:: Expected: "
                    f"{project.expected_errored}, Actual: {report.number_errored}"
                )

            exit_code = (
                exit_code
                or report.number_changed != project.expected_changed
                or report.number_unchanged != project.expected_unchanged
                or report.number_errored != project.expected_errored
            )

    click.echo(f"Exiting with code {exit_code:d}", err=True)
    ctx.exit(exit_code)


def get_project_source_tree(
    project: SQLProject, reset_cache: bool, working_dir: str
) -> Path:
    """
    Returns a Path to a directory containing a project's source tree. Defaults to
    using a cached copy, but will clone a repo and checkout a specific ref if the
    cache does not exist or if reset_cache is passed (in this case, it copies the
    newly-checked-out tree to the cache)
    """

    cache_dir = Path(user_cache_dir(appname="sqlfmt_primer"))
    proj_cache_dir = cache_dir / project.name / project.git_ref

    if reset_cache or not proj_cache_dir.exists():
        click.echo(f"Cloning {project.name}", err=True)
        target_dir = clone_and_checkout(project, working_dir)
        shutil.copytree(target_dir, proj_cache_dir, dirs_exist_ok=True)
        return target_dir
    else:
        click.echo(f"Using cached files for {project.name}", err=True)
        return proj_cache_dir


def clone_and_checkout(project: SQLProject, working_dir: str) -> Path:
    """
    Clone a project into a local dir, and checkout its ref.
    Returns the local directory containing the source tree
    """
    repo_dir = Path(working_dir) / project.name
    repo = Repo.clone_from(url=project.git_url, to_path=repo_dir)
    head = repo.create_head("sqlfmt", commit=project.git_ref)
    head.checkout(force=True)
    target_dir = repo_dir / project.sub_directory if project.sub_directory else repo_dir
    assert target_dir.exists(), f"Failed to clone repo {project.name}"
    return target_dir


def clear_sqlfmt_cache() -> None:
    """
    Deletes the cache file from the disk, if it exists. Called before
    each primer run to ensure we're formatting every file every time.
    """
    p = get_cache_file()
    try:
        p.unlink()
    except FileNotFoundError:
        pass


def _warn(msg: str) -> None:
    """
    Make msg bold and yellow and print to stderr
    """
    click.echo(
        click.style(msg, fg="yellow", bold=True),
        err=True,
    )



class SqlfmtError(ValueError):
    """
    Generic class for an error encountered when formatting
    user code. Should be caught by api.run for display to the
    user.
    """

    def __str__(self) -> str:
        intro = "sqlfmt encountered an error: "
        message = super().__str__()
        return f"{intro}{message}"


class SqlfmtConfigError(SqlfmtError):
    """
    Raised during opening or parsing a pyproject.toml file
    """

    pass


class SqlfmtUnicodeError(SqlfmtError):
    """
    Raised while reading input if the input cannot be
    decoded into a Python string
    """

    pass


class SqlfmtParsingError(SqlfmtError):
    """
    Raised during lexing if sqlfmt encounters a token that does
    not match any rules
    """

    pass


class SqlfmtBracketError(SqlfmtError):
    """
    Raised during lexing if sqlfmt encounters mismatched, unterminated,
    or unexpected closing brackets
    """

    pass


class SqlfmtSegmentError(SqlfmtError):
    """
    Raised during merging if a segment is unexpectedly empty
    """

    pass


class SqlfmtEquivalenceError(SqlfmtError):
    """
    Raised during the safety check if the result query does
    not lex to the same tokens as the raw query
    """

    pass


class SqlfmtControlFlowException(Exception):
    """
    Generic exception for exceptions used to manage control
    flow. Should always be caught by the application and
    never raised back to the user
    """

    pass


class StopRulesetLexing(SqlfmtControlFlowException):
    """
    Raised by the Analyzer or one of its actions to indicate
    that further lexing should use the previous ruleset
    """

    pass


class CannotMergeException(SqlfmtControlFlowException):
    """
    Raised by the merger if the passed lines cannot be merged
    for any reason
    """

    pass



import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from sqlfmt.dialect import ClickHouse, Polyglot
from sqlfmt.exception import SqlfmtConfigError


@dataclass
class Mode:
    """
    A Mode is a container for all sqlfmt config, including formatting config and
    report config. For more info on each option, see cli.py
    """

    SQL_EXTENSIONS: List[str] = field(default_factory=lambda: [".sql", ".sql.jinja"])
    dialect_name: str = "polyglot"
    line_length: int = 88
    check: bool = False
    diff: bool = False
    exclude: List[str] = field(default_factory=list)
    exclude_root: Optional[Path] = None
    encoding: str = "utf-8"
    fast: bool = False
    single_process: bool = False
    no_jinjafmt: bool = False
    reset_cache: bool = False
    verbose: bool = False
    quiet: bool = False
    no_progressbar: bool = False
    no_color: bool = False
    force_color: bool = False

    def __post_init__(self) -> None:
        # get the dialect from its name.
        dialects = {
            "polyglot": Polyglot,
            "clickhouse": ClickHouse,
        }
        try:
            self.dialect = dialects[self.dialect_name.lower()]()
        except KeyError as e:
            raise SqlfmtConfigError(
                f"Mode was created with dialect_name={self.dialect_name}, "
                "which is not supported. Did you mean 'polyglot'?"
            ) from e

    @property
    def color(self) -> bool:
        """
        There are 4 considerations for setting color:
        1. The --force-color option
        2. The --no-color option
        3. The NO_COLOR environment variable
        4. The default behavior, which is to colorize output

        This property checks these flags, in descending order of priority,
        and sets the authoritative flag accordingly

        See no-color.org for details.
        """
        if self.force_color:
            return True
        elif self.no_color:
            return False
        elif os.environ.get("NO_COLOR", False):
            return False
        else:
            return True



from pathlib import Path

import pytest
from click.testing import CliRunner

from sqlfmt.cli import sqlfmt as sqlfmt_main


@pytest.mark.parametrize(
    "options",
    [
        "",
        "--verbose",
        "--line-length 88",
        "-l 88",
        "--check",
        "--check -v",
        "--diff",
        "-v --diff",
        "--check --diff",
        "--no-color",
        "--diff --no-color",
        "--no-jinjafmt",
        "-k",
        "--reset-cache",
        "--dialect clickhouse",
    ],
)
def test_end_to_end_preformatted(
    sqlfmt_runner: CliRunner, preformatted_dir: Path, options: str
) -> None:
    # CliRunner removes backslashes from strings passed in as args,
    # which makes this test fail when running on windows. We need
    # to format the path as a posix path (with forward slashes)
    # as a workaround.
    args = f"{preformatted_dir.as_posix()} {options}"
    result = sqlfmt_runner.invoke(sqlfmt_main, args=args)

    preformatted_count = sum((1 for f in preformatted_dir.iterdir() if f.is_file()))

    assert result
    assert f"{preformatted_count} files" in result.stderr

    if "check" in options:
        assert "passed formatting check" in result.stderr
    elif "diff" in options:
        assert "would be left unchanged" in result.stderr
    else:
        assert "left unchanged" in result.stderr

    if "-v" in options or "--verbose" in options:
        assert "001_select_1.sql" in result.stderr

    assert result.exit_code == 0


@pytest.mark.parametrize(
    "options",
    ["-q", "--quiet", "--check -q", "--diff -q"],
)
def test_end_to_end_preformatted_quiet(
    sqlfmt_runner: CliRunner, preformatted_dir: Path, options: str
) -> None:
    # CliRunner removes backslashes from strings passed in as args,
    # which makes this test fail when running on windows. We need
    # to format the path as a posix path (with forward slashes)
    # as a workaround.
    args = f"{preformatted_dir.as_posix()} {options}"
    result = sqlfmt_runner.invoke(sqlfmt_main, args=args)

    assert result
    assert result.stderr == ""
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "options",
    [
        "--check",
        "--check -v",
        "--diff",
        "-v --diff",
        "--diff --check",
        "--check --no-color",
        "--diff --no-color",
        "--dialect clickhouse --check",
    ],
)
def test_end_to_end_check_unformatted(
    sqlfmt_runner: CliRunner, unformatted_dir: Path, options: str
) -> None:
    # CliRunner removes backslashes from strings passed in as args,
    # which makes this test fail when running on windows. We need
    # to format the path as a posix path (with forward slashes)
    # as a workaround.
    args = f"{unformatted_dir.as_posix()} {options}"
    result = sqlfmt_runner.invoke(sqlfmt_main, args=args)

    unformatted_count = sum((1 for f in unformatted_dir.iterdir() if f.is_file()))

    assert result
    assert f"{unformatted_count} files" in result.stderr
    if "diff" in options and "check" not in options:
        assert "would be formatted" in result.stderr
    else:
        assert "failed formatting check" in result.stderr

    assert result.exit_code == 1


@pytest.mark.parametrize(
    "options",
    [
        "-q --check",
        "--quiet --check",
        "--check -q",
    ],
)
def test_end_to_end_check_unformatted_quiet(
    sqlfmt_runner: CliRunner, unformatted_dir: Path, options: str
) -> None:
    # CliRunner removes backslashes from strings passed in as args,
    # which makes this test fail when running on windows. We need
    # to format the path as a posix path (with forward slashes)
    # as a workaround.
    args = f"{unformatted_dir.as_posix()} {options}"
    result = sqlfmt_runner.invoke(sqlfmt_main, args=args)

    assert result
    assert result.stderr == ""
    assert result.exit_code == 1


@pytest.mark.parametrize(
    "options", ["", "--check", "--no-color", "--quiet", "--dialect clickhouse"]
)
def test_end_to_end_errors(
    sqlfmt_runner: CliRunner, error_dir: Path, options: str
) -> None:
    # CliRunner removes backslashes from strings passed in as args,
    # which makes this test fail when running on windows. We need
    # to format the path as a posix path (with forward slashes)
    # as a workaround.
    args = f"{error_dir.as_posix()} {options}"
    result = sqlfmt_runner.invoke(sqlfmt_main, args=args)

    error_count = sum((1 for f in error_dir.iterdir() if f.is_file()))

    assert result
    if "quiet" not in options:
        assert f"{error_count} files had errors" in result.stderr
    assert "sqlfmt encountered an error" in result.stderr
    assert "900_bad_token.sql" in result.stderr
    assert result.exit_code == 2


@pytest.mark.parametrize(
    "options,stream_input,expected_stdout,expected_exit",
    [
        ("-", "select    1\n", "select 1\n", 0),  # removes excess space
        ("-", "select 1\n\n", "select 1\n", 0),  # one trailing newline
        ("- --check", "select    1\n", "", 1),
        ("- --diff", "select    1\n", "", 1),
        ("- --check", "select 1\n", "", 0),
        ("- --diff", "select 1\n", "", 0),
    ],
)
def test_stdin(
    sqlfmt_runner: CliRunner,
    options: str,
    stream_input: str,
    expected_stdout: str,
    expected_exit: int,
) -> None:
    results = sqlfmt_runner.invoke(sqlfmt_main, args=options, input=stream_input)
    assert results.exit_code == expected_exit
    assert results.stdout == expected_stdout



import os
import sys
from itertools import product
from pathlib import Path
from typing import Any

import pytest
from click.testing import CliRunner

from sqlfmt.analyzer import Analyzer
from sqlfmt.mode import Mode
from tests.util import copy_test_data_to_tmp

# make tests module importable
sys.path.append(os.path.dirname(__file__))


def pytest_sessionstart(session: pytest.Session) -> None:
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    from sqlfmt.cache import clear_cache
    from tests.util import delete_results_dir

    delete_results_dir()
    clear_cache()


@pytest.fixture
def sqlfmt_runner() -> CliRunner:
    return CliRunner()


@pytest.fixture
def unset_no_color_env(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("NO_COLOR", raising=False)


@pytest.fixture
def default_mode(unset_no_color_env: None) -> Mode:
    return Mode()


@pytest.fixture
def clickhouse_mode(unset_no_color_env: None) -> Mode:
    return Mode(dialect_name="clickhouse")


@pytest.fixture
def verbose_mode(unset_no_color_env: None) -> Mode:
    return Mode(verbose=True)


@pytest.fixture
def check_mode(unset_no_color_env: None) -> Mode:
    return Mode(check=True)


@pytest.fixture
def verbose_check_mode(unset_no_color_env: None) -> Mode:
    return Mode(check=True, verbose=True)


@pytest.fixture
def diff_mode(unset_no_color_env: None) -> Mode:
    return Mode(diff=True)


@pytest.fixture
def no_color_diff_mode(unset_no_color_env: None) -> Mode:
    return Mode(diff=True, no_color=True)


@pytest.fixture
def reset_cache_mode(unset_no_color_env: None) -> Mode:
    return Mode(reset_cache=True)


@pytest.fixture
def no_progressbar_mode(unset_no_color_env: None) -> Mode:
    return Mode(no_progressbar=True)


@pytest.fixture
def single_process_mode(unset_no_color_env: None) -> Mode:
    return Mode(single_process=True)


@pytest.fixture(params=product([True, False], repeat=4))
def all_output_modes(request: Any, unset_no_color_env: None) -> Mode:
    return Mode(
        check=request.param[0],
        diff=request.param[1],
        single_process=request.param[2],
        no_progressbar=request.param[3],
    )


@pytest.fixture
def default_analyzer(default_mode: Mode) -> Analyzer:
    return default_mode.dialect.initialize_analyzer(default_mode.line_length)


@pytest.fixture
def preformatted_dir(tmp_path: Path) -> Path:
    """
    Copies the directory of preformatted sql files from the unit test
    directory into a temp directory (provided by pytest fixture tmp_path),
    and then returns the path to the temp directory.
    """
    test_dir = copy_test_data_to_tmp(["fast/preformatted"], tmp_path)
    return test_dir


@pytest.fixture
def unformatted_dir(tmp_path: Path) -> Path:
    """
    Copies the directory of unformatted sql files from the test/data
    directory into a temp directory (provided by pytest fixture tmp_path),
    and then returns the path to the temp directory.
    """
    test_dir = copy_test_data_to_tmp(["fast/unformatted"], tmp_path)
    return test_dir


@pytest.fixture
def error_dir(tmp_path: Path) -> Path:
    test_dir = copy_test_data_to_tmp(["fast/errors"], tmp_path)
    return test_dir



import itertools

import pytest

from sqlfmt.exception import CannotMergeException
from sqlfmt.merger import LineMerger
from sqlfmt.mode import Mode
from sqlfmt.operator_precedence import OperatorPrecedence
from sqlfmt.segment import Segment, create_segments_from_lines
from tests.util import read_test_data


@pytest.fixture
def merger(default_mode: Mode) -> LineMerger:
    return LineMerger(default_mode)


def test_create_merged_line(merger: LineMerger) -> None:
    source_string = """
    select
        able,
        baker,




        charlie,





        delta,
    """
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)

    expected = "\nselect able, baker,\n"
    merged = merger.create_merged_line(raw_query.lines[0:4])
    actual = "".join([str(line) for line in merged])
    assert actual == expected

    expected = "\nselect able, baker, charlie, delta,\n"
    merged = merger.create_merged_line(raw_query.lines)
    actual = "".join([str(line) for line in merged])
    assert actual == expected

    with pytest.raises(CannotMergeException):
        # can't merge whitespace
        _ = merger.create_merged_line(raw_query.lines[-6:-3])


def test_create_merged_line_comments(merger: LineMerger) -> None:
    source_string = """
    select
        able,
        baker,
        -- a comment
        charlie,
    """
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)

    with pytest.raises(CannotMergeException) as exc_info:
        # can't merge whitespace
        _ = merger.create_merged_line(raw_query.lines)

    assert "comment" in str(exc_info.value)


def test_basic_merge(merger: LineMerger) -> None:
    source_string = """
        nullif(
            full_name,
            ''
        ) as c,
    """.strip()
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    assert len(merged_lines) == 1

    expected_string = "nullif(full_name, '') as c,"

    result = "".join(map(str, merged_lines)).strip()
    assert result == expected_string


def test_nested_merge(merger: LineMerger) -> None:
    source_string = """
    select
        a,
        nullif(
            split_part(
                full_name,
                ' ',
                2
            ),
            ''
        ) as last_name,
    """
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)

    expected_string = (
        "select a, nullif(split_part(full_name, ' ', 2), '') as last_name,"
    )

    result = "".join(map(str, merged_lines)).strip()
    assert result == expected_string


def test_incomplete_merge(merger: LineMerger) -> None:
    source_string = """
    select
        first_field,
        nullif(
            split_part(
                full_name,
                ' ',
                2
            ),
            ''
        ) as last_name,
        another_field,
        case
            when short
            then 1
        end,
        yet_another_field,
        and_still_another_field
    from
        my_table
    where
        some_condition is true
    """
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)

    result = list(map(str, merged_lines))

    expected = [
        "\n",
        "select\n",
        "    first_field,\n",
        "    nullif(split_part(full_name, ' ', 2), '') as last_name,\n",
        "    another_field,\n",
        "    case when short then 1 end,\n",
        "    yet_another_field,\n",
        "    and_still_another_field\n",
        "from my_table\n",
        "where some_condition is true\n",
    ]

    assert result == expected


def test_cte_merge(merger: LineMerger) -> None:
    source_string = """
    with
        my_cte as (
            select * from my_table
        )
    select * from my_cte
    """.strip()
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)

    result = list(map(str, merged_lines))

    expected = [
        "with my_cte as (select * from my_table) select * from my_cte\n",
    ]

    assert result == expected


def test_case_then_merge(merger: LineMerger) -> None:
    source_string = """
    case
        when
            some_initial_condition_is_true
        then
            some_other_condition
        else
            something_else_entirely
    end
    """
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)

    result = list(map(str, merged_lines))

    expected = [
        "\n",
        "case\n",
        "    when some_initial_condition_is_true\n",
        "    then some_other_condition\n",
        "    else something_else_entirely\n",
        "end\n",
    ]

    assert result == expected


def test_merge_count_window_function(merger: LineMerger) -> None:
    source_string = (
        "count(\n"
        "    case\n"
        "        when\n"
        "            a\n"
        "            is null\n"
        "        then\n"
        "            1\n"
        "    end\n"
        ")\n"
        "over (\n"
        "    partition by\n"
        "        user_id,\n"
        "        date_trunc(\n"
        "            'year',\n"
        "            performed_at\n"
        "        )\n"
        ")\n"
        "as d,\n"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)

    result = list(map(str, merged_lines))

    expected = [
        "count(case when a is null then 1 end) over (\n",
        "    partition by user_id, date_trunc('year', performed_at)\n",
        ") as d,\n",
    ]

    assert result == expected


def test_disallow_multiline(merger: LineMerger) -> None:
    source_string = """
        foo
        {{
            bar
        }}
    """.strip()
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    with pytest.raises(CannotMergeException):
        _ = merger.create_merged_line(raw_query.lines)


def test_segment_continues_operator_sequence(merger: LineMerger) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_merger/test_segment_continues_operator_sequence.sql"
    )
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )

    segments = create_segments_from_lines(q.lines)
    assert len(segments) == 9
    print([str(s[0]) for s in segments])

    p_ON_result = [
        merger._segment_continues_operator_sequence(
            s, max_precedence=OperatorPrecedence.ON
        )
        for s in segments
    ]
    p_ON_expected = [False, True, True, True, False, True, False, True, True]
    assert p_ON_result == p_ON_expected

    p_PRESENCE_result = [
        merger._segment_continues_operator_sequence(
            s, max_precedence=OperatorPrecedence.PRESENCE
        )
        for s in segments
    ]
    p_PRESENCE_expected = [False, True, True, False, False, True, False, False, True]
    assert p_PRESENCE_result == p_PRESENCE_expected

    p_MULTIPLICATION_result = [
        merger._segment_continues_operator_sequence(
            s, max_precedence=OperatorPrecedence.MULTIPLICATION
        )
        for s in segments
    ]
    p_MULTIPLICATION_expected = [
        False,
        False,
        True,
        False,
        False,
        True,
        False,
        False,
        False,
    ]
    assert p_MULTIPLICATION_result == p_MULTIPLICATION_expected

    p_OTHER_TIGHT_result = [
        merger._segment_continues_operator_sequence(
            s, max_precedence=OperatorPrecedence.OTHER_TIGHT
        )
        for s in segments
    ]
    p_OTHER_TIGHT_expected = [
        False,
        False,
        False,
        False,
        False,
        True,
        False,
        False,
        False,
    ]
    assert p_OTHER_TIGHT_result == p_OTHER_TIGHT_expected


@pytest.mark.parametrize("p", OperatorPrecedence.tiers())
def test_segment_continues_operator_sequence_empty(
    merger: LineMerger, p: OperatorPrecedence
) -> None:
    source_string = "\n\n\n\n"
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )
    result = merger._segment_continues_operator_sequence(
        Segment(q.lines), max_precedence=p
    )
    assert result is True


def test_merge_single_line(merger: LineMerger) -> None:
    source_string = "select 1\n"
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )
    merged_lines = merger.create_merged_line(q.lines)
    assert merged_lines == q.lines


def test_merge_lines_split_by_operators(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_merge_lines_split_by_operators.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string


def test_merge_chained_parens(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_merge_chained_parens.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string


def test_merge_operators_before_children(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_merge_operators_before_children.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string


def test_do_not_merge_very_long_chains(merger: LineMerger) -> None:
    source_string = "a" + ("\n+ b" * 40) + "\n"
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])

    assert result_string == source_string


def test_respect_extra_blank_lines(merger: LineMerger) -> None:
    source_string = """
    select
        one_field,
        another_field,
        yet_another_field


    from
        my_table

    join
        your_table
        on my_table.your_id
        = your_table.my_id


    where
        something is true
    """.lstrip()
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    assert merged_lines[1].is_blank_line
    assert merged_lines[2].is_blank_line
    assert merged_lines[4].is_blank_line
    assert merged_lines[6].is_blank_line
    assert merged_lines[7].is_blank_line


def test_maybe_merge_blank_lines(merger: LineMerger) -> None:
    source_string = "select\n\n\n\n\n\n\n\n\n\n\n\n1\n"
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    blank_lines = raw_query.lines[1:-1]
    assert all(line.is_blank_line for line in blank_lines)
    merged_lines = merger.maybe_merge_lines(blank_lines)
    assert merged_lines == blank_lines


def test_stubborn_merge_blank_lines(merger: LineMerger) -> None:
    source_string = "a\n\n\n\n\n\n\n\n\n\n\n\nb\n"
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    prev_segment = Segment(raw_query.lines[:1])
    blank_segment = Segment(raw_query.lines[1:-1])
    assert all(line.is_blank_line for line in blank_segment)
    new_segments = merger._stubbornly_merge(
        prev_segments=[prev_segment], segment=blank_segment
    )
    assert new_segments == [prev_segment, blank_segment]


@pytest.mark.parametrize("sep", [";", "union", "union all", "intersect", "except"])
def test_do_not_merge_across_query_dividers(merger: LineMerger, sep: str) -> None:
    source_string = f"select 1\n{sep}\nselect2\n"
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    assert raw_query.lines == merged_lines


def test_maybe_stubbornly_merge(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_maybe_stubbornly_merge.sql"
    )
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )
    segments = create_segments_from_lines(q.lines)
    merged_segments = merger._maybe_stubbornly_merge(segments)
    result_string = "".join(
        [line.render_with_comments(88) for line in itertools.chain(*merged_segments)]
    )
    assert result_string == expected_string


def test_maybe_stubbornly_merge_single_segment(merger: LineMerger) -> None:
    source_string = "select\na,\nb\n"
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )
    segments = create_segments_from_lines(q.lines)
    assert len(segments) == 1
    merged_segments = merger._maybe_stubbornly_merge(segments)
    assert merged_segments == segments


def test_fix_standalone_operators(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_fix_standalone_operators.sql"
    )
    q = merger.mode.dialect.initialize_analyzer(merger.mode.line_length).parse_query(
        source_string
    )

    # first pass shouldn't do anything
    segments = create_segments_from_lines(q.lines)
    assert len(segments) == 11
    fixed_segments = merger._fix_standalone_operators(segments)
    assert len(fixed_segments) == len(segments)

    result_string = "".join([str(line) for line in itertools.chain(*fixed_segments)])
    assert result_string == expected_string


def test_merge_inline_comments(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_merge_inline_comments.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([line.render_with_comments(88) for line in merged_lines])
    assert result_string == expected_string


def test_no_merge_short_multiline_nodes(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_no_merge_short_multiline_nodes.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([line.render_with_comments(88) for line in merged_lines])
    assert result_string == expected_string


def test_no_merge_formatting_disabled(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_no_merge_formatting_disabled.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string


def test_do_not_merge_operator_sequences_across_commas(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_no_merge_operator_sequences_across_commas.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string


def test_do_not_merge_databricks_query_hints(merger: LineMerger) -> None:
    source_string, expected_string = read_test_data(
        "unit_tests/test_merger/test_no_merge_databricks_query_hints.sql"
    )
    raw_query = merger.mode.dialect.initialize_analyzer(
        merger.mode.line_length
    ).parse_query(source_string)
    merged_lines = merger.maybe_merge_lines(raw_query.lines)
    result_string = "".join([str(line) for line in merged_lines])
    assert result_string == expected_string



from dataclasses import dataclass, field
from typing import List, Optional, Tuple

from sqlfmt.tokens import Token, TokenType


def get_previous_token(prev_node: Optional["Node"]) -> Tuple[Optional[Token], bool]:
    """
    Returns the token of prev_node, unless prev_node is a
    newline or jinja statement, in which case it recurses
    """
    if not prev_node:
        return None, False
    t = prev_node.token
    if t.type.does_not_set_prev_sql_context:
        prev, _ = get_previous_token(prev_node.previous_node)
        return prev, True
    else:
        return t, False


@dataclass
class Node:
    """
    A Node wraps a lexed Token, but adds many calculated properties and methods that
    simplify formatting, including:

    previous_node: a reference to the Node that immediately precedes this Node in the
    query

    prefix: the calculated whitespace (0 or 1 spaces) that should precede this Node
    when formatted

    value: the properly-capitalized token contents for the formatted query

    open_brackets and open_jinja_blocks: a list of Nodes that precede this Node that
    refer to open brackets (keywords and parens) or jinja blocks (e.g., {% if foo %})
    that increase the syntax depth (and therefore printed indentation) of this Node

    formatting_disabled: a list of FMT_OFF tokens that precede this node and prevent
    it from being formatted
    """

    token: Token
    previous_node: Optional["Node"]
    prefix: str
    value: str
    open_brackets: List["Node"] = field(default_factory=list)
    open_jinja_blocks: List["Node"] = field(default_factory=list)
    formatting_disabled: List[Token] = field(default_factory=list)

    def __str__(self) -> str:
        """
        Returns the formatted text of this Node
        """
        return f"{self.prefix}{self.value}"

    def __repr__(self) -> str:
        """
        Because of self.previous_node, the default dataclass repr creates
        unusable output
        """

        def simple_node(node: Optional[Node]) -> str:
            return f"Node(token={node.token})" if node else "None"

        prev = simple_node(self.previous_node)
        b = [simple_node(n) for n in self.open_brackets]
        j = [simple_node(n) for n in self.open_jinja_blocks]
        r = (
            f"Node(\n"
            f"\ttoken='{str(self.token)}',\n"
            f"\tprevious_node={prev},\n"
            f"\tdepth={self.depth},\n"
            f"\tprefix='{self.prefix}',\n"
            f"\tvalue='{self.value}',\n"
            f"\topen_brackets={b},\n"
            f"\topen_jinja_blocks={j},\n"
            f"\tformatting_disabled={self.formatting_disabled}\n"
            f")"
        )
        return r

    def __len__(self) -> int:
        """
        The length of this printed Node, including prefix whitespace, after formatting
        """
        return len(str(self))

    @property
    def depth(self) -> Tuple[int, int]:
        """
        A Node's depth is a key characteristic that determines its indentation in the
        formatted query. We use a tuple to track SQL and jinja depth separately, since
        SQL depth can change within jinja blocks
        """
        return (len(self.open_brackets), len(self.open_jinja_blocks))

    @property
    def is_unterm_keyword(self) -> bool:
        """
        True for Nodes representing unterminated SQL keywords, like select, from, where
        """
        return self.token.type is TokenType.UNTERM_KEYWORD

    @property
    def is_comma(self) -> bool:
        return self.token.type is TokenType.COMMA

    @property
    def divides_queries(self) -> bool:
        return self.token.type.divides_queries

    @property
    def is_opening_bracket(self) -> bool:
        return self.token.type.is_opening_bracket

    @property
    def is_bracket_operator(self) -> bool:
        """
        Node is an opening square bracket ("[")
        that follows a token that could be a name.

        Alternatively, node is an open paren ("(")
        that follow an closing angle bracket.
        """
        if self.token.type is not TokenType.BRACKET_OPEN:
            return False

        prev_token, _ = get_previous_token(self.previous_node)
        if not prev_token:
            return False
        elif self.value == "[":
            return prev_token.type in (
                TokenType.NAME,
                TokenType.QUOTED_NAME,
                TokenType.BRACKET_CLOSE,
            )
        # BQ struct literals have parens that follow closing angle
        # brackets
        else:
            return (
                self.value == "("
                and prev_token.type is TokenType.BRACKET_CLOSE
                and ">" in prev_token.token
            )

    @property
    def is_closing_bracket(self) -> bool:
        return self.token.type in (
            TokenType.BRACKET_CLOSE,
            TokenType.STATEMENT_END,
        )

    @property
    def is_opening_jinja_block(self) -> bool:
        return self.token.type in (
            TokenType.JINJA_BLOCK_START,
            TokenType.JINJA_BLOCK_KEYWORD,
        )

    @property
    def is_jinja(self) -> bool:
        return self.token.type.is_jinja

    @property
    def is_closing_jinja_block(self) -> bool:
        return self.token.type is TokenType.JINJA_BLOCK_END

    @property
    def is_jinja_block_keyword(self) -> bool:
        return self.token.type is TokenType.JINJA_BLOCK_KEYWORD

    @property
    def is_jinja_statement(self) -> bool:
        return self.token.type.is_jinja_statement

    @property
    def is_operator(self) -> bool:
        return (
            self.token.type.is_always_operator
            or self.is_multiplication_star
            or self.is_bracket_operator
        )

    @property
    def is_boolean_operator(self) -> bool:
        return self.token.type is TokenType.BOOLEAN_OPERATOR

    @property
    def is_multiplication_star(self) -> bool:
        """
        A lexed TokenType.STAR token can be the "all fields" shorthand or
        the multiplication operator. Returns true iff this Node is a multiplication
        operator
        """
        if self.token.type is not TokenType.STAR:
            return False
        prev_token, _ = get_previous_token(self.previous_node)
        if not prev_token:
            return False
        else:
            return prev_token.type not in (
                TokenType.UNTERM_KEYWORD,
                TokenType.COMMA,
                TokenType.DOT,
            )

    @property
    def is_the_between_operator(self) -> bool:
        """
        True if this node is a WORD_OPERATOR with the value "between"
        """
        return self.token.type is TokenType.WORD_OPERATOR and self.value == "between"

    @property
    def has_preceding_between_operator(self) -> bool:
        """
        True if this node has a preceding "between" operator at the same depth
        """
        prev = (
            self.previous_node.previous_node if self.previous_node is not None else None
        )
        while prev and prev.depth >= self.depth:
            if prev.depth == self.depth and prev.is_the_between_operator:
                return True
            elif prev.depth == self.depth and prev.is_boolean_operator:
                break
            else:
                prev = prev.previous_node
        return False

    @property
    def is_the_and_after_the_between_operator(self) -> bool:
        """
        True if this node is a BOOLEAN_OPERATOR with the value "and" immediately
        following a "between" operator
        """
        if not self.is_boolean_operator or self.value != "and":
            return False
        else:
            return self.has_preceding_between_operator

    @property
    def is_newline(self) -> bool:
        return self.token.type is TokenType.NEWLINE

    @property
    def is_multiline_jinja(self) -> bool:
        if self.token.type.is_jinja and "\n" in self.value:
            return True
        else:
            return False



import itertools
from dataclasses import dataclass
from typing import Iterable, List, Optional, Tuple

from sqlfmt.comment import Comment
from sqlfmt.exception import CannotMergeException, SqlfmtSegmentError
from sqlfmt.line import Line
from sqlfmt.mode import Mode
from sqlfmt.node import Node
from sqlfmt.operator_precedence import OperatorPrecedence
from sqlfmt.segment import Segment, create_segments_from_lines


@dataclass
class LineMerger:
    mode: Mode

    def create_merged_line(self, lines: List[Line]) -> List[Line]:
        """
        Returns a new line by merging together all nodes in lines. Raises an
        exception if the returned line would be too long, empty, or the nodes in
        any of the lines violate the rules in _raise_unmergeable.
        """

        if len(lines) <= 1:
            return lines

        nodes, comments = self._extract_components(lines)

        merged_line = Line.from_nodes(
            previous_node=lines[0].previous_node,
            nodes=nodes,
            comments=comments,
        )

        if merged_line.is_too_long(self.mode.line_length):
            raise CannotMergeException("Merged line is too long")

        # add in any leading or trailing blank lines
        leading_blank_lines = self._extract_leading_blank_lines(lines)
        trailing_blank_lines = list(
            reversed(self._extract_leading_blank_lines(reversed(lines)))
        )

        return leading_blank_lines + [merged_line] + trailing_blank_lines

    def safe_create_merged_line(self, lines: List[Line]) -> List[Line]:
        try:
            return self.create_merged_line(lines)
        except CannotMergeException:
            return lines

    @classmethod
    def _extract_components(
        cls, lines: Iterable[Line]
    ) -> Tuple[List[Node], List[Comment]]:
        """
        Given a list of lines, return 2 components:
        1. list of all nodes in those lines, with only a single trailing newline
        2. list of all comments in all of those lines

        Raise CannotMergeException if lines contain nodes that cannot
        be merged.
        """
        nodes: List[Node] = []
        comments: List[Comment] = []
        final_newline: Optional[Node] = None
        allow_multiline_jinja = True
        has_multiline_jinja = False
        has_inline_comment_above = False
        for line in lines:
            # only merge lines with comments if it's a standalone comment
            # above the first line or an inline comment after the last
            # line
            if line.comments:
                if has_inline_comment_above:
                    raise CannotMergeException(
                        "Can't merge lines with inline comments and other comments"
                    )
                elif any(
                    [comment.is_databricks_query_hint for comment in line.comments]
                ):
                    raise CannotMergeException(
                        "Can't merge lines with a databricks type hint comment"
                    )
                elif (
                    len(line.comments) == 1
                    and len(line.nodes) > 1
                    and (
                        line.comments[0].is_inline
                        or line.comments[0].previous_node == line.nodes[-2]
                    )
                ):
                    # this is a comment that must be rendered inline,
                    # so it'll probably block merging unless the
                    # next line is just a comma
                    has_inline_comment_above = True
                elif len(nodes) == 1 and nodes[0].is_operator:
                    # if source has standalone operators, we can merge
                    # the operator into the contents, even if there is
                    # a comment in the way
                    pass
                elif nodes:
                    raise CannotMergeException(
                        "Can't merge lines with standalone comments unless the "
                        "comments are above the first line"
                    )
            # make an exception for inline comments followed by
            # a lonely comma (e.g., leading commas with inline comments)
            elif has_inline_comment_above:
                if not (line.is_standalone_comma or line.is_blank_line):
                    raise CannotMergeException(
                        "Can't merge lines with inline comments unless "
                        "the following line is a single standalone comma "
                        "or a blank line"
                    )

            if has_multiline_jinja and not (
                line.starts_with_operator or line.starts_with_comma
            ):
                raise CannotMergeException(
                    "Can't merge lines containing multiline nodes"
                )
            # skip over newline nodes
            content_nodes = [
                cls._raise_unmergeable(node, allow_multiline_jinja)
                for node in line.nodes
                if not node.is_newline
            ]
            if content_nodes:
                final_newline = line.nodes[-1]
                nodes.extend(content_nodes)
                # we can merge a line containing multiline jinja
                # into a preceding line iff:
                # the multiline node is on the second line and follows a
                # standalone operator
                if not (
                    allow_multiline_jinja
                    and len(content_nodes) == 1
                    and content_nodes[0].is_operator
                ):
                    allow_multiline_jinja = False
                # we can merge a line into a preceding line that
                # contains multiline jinja iff:
                # the line starts with an operator or a comma
                has_multiline_jinja = any(
                    [node.is_multiline_jinja for node in content_nodes]
                )
            comments.extend(line.comments)

        if not nodes or not final_newline:
            raise CannotMergeException("Can't merge only whitespace/newlines")

        nodes.append(final_newline)

        return nodes, comments

    @staticmethod
    def _raise_unmergeable(node: Node, allow_multiline_jinja: bool) -> Node:
        """
        Raises a CannotMergeException if the node cannot be merged. Otherwise
        returns the node
        """
        if node.formatting_disabled:
            raise CannotMergeException(
                "Can't merge lines containing disabled formatting"
            )
        elif node.divides_queries:
            raise CannotMergeException(
                "Can't merge multiple queries onto a single line"
            )
        elif node.is_multiline_jinja and not allow_multiline_jinja:
            raise CannotMergeException("Can't merge lines containing multiline nodes")
        else:
            return node

    @staticmethod
    def _extract_leading_blank_lines(lines: Iterable[Line]) -> List[Line]:
        leading_blank_lines: List[Line] = []
        for line in lines:
            if line.is_blank_line:
                leading_blank_lines.append(line)
            else:
                break
        return leading_blank_lines

    def maybe_merge_lines(self, lines: List[Line]) -> List[Line]:
        """
        Tries to merge lines into a single line; if that fails,
        splits lines into segments of equal depth, merges
        runs of operators at that depth, and then recurses into
        each segment

        Returns a new list of Lines
        """
        if not lines or all([line.formatting_disabled for line in lines]):
            return lines

        try:
            merged_lines = self.create_merged_line(lines)
        except CannotMergeException:
            merged_lines = []
            # doesn't fit onto a single line, so split into
            # segments at the depth of lines[0]
            segments = create_segments_from_lines(lines)
            # if a segment starts with a standalone operator,
            # the first two lines of that segment should likely
            # be merged before doing anything else
            segments = self._fix_standalone_operators(segments)
            if len(segments) > 1:
                # merge together segments of equal depth that are
                # joined by operators
                segments = self._maybe_merge_operators(
                    segments, OperatorPrecedence.tiers()
                )
                # some operators really should not be by themselves
                # so if their segments are too long to be merged,
                # we merge just their first line onto the prior segment
                segments = self._maybe_stubbornly_merge(segments)
                # then recurse into each segment and try to merge lines
                # within individual segments
                for segment in segments:
                    merged_lines.extend(self.maybe_merge_lines(segment))
            # if there was only a single segment at the depth of the
            # top line, we need to move down one line and try again.
            # Because of the structure of a well-split set of lines,
            # in this case moving down one line is guaranteed to move
            # us in one depth.
            # if the final line of the segment matches the top line,
            # we need to strip that off so we only segment the
            # indented lines
            else:
                only_segment = segments[0]
                try:
                    _, i = only_segment.head
                except SqlfmtSegmentError:
                    merged_lines.extend(only_segment)
                else:
                    merged_lines.extend(only_segment[: i + 1])
                    for segment in only_segment.split_after(i):
                        merged_lines.extend(self.maybe_merge_lines(segment))

        return merged_lines

    def _fix_standalone_operators(self, segments: List[Segment]) -> List[Segment]:
        """
        If the first line of a segment is a standalone operator,
        we should try to merge the first two lines together before
        doing anything else
        """
        for segment in segments:
            try:
                head, i = segment.head
                if head.is_standalone_operator:
                    remainder_after_operator = Segment(segment[i + 1 :])
                    _, j = remainder_after_operator.head
                    try:
                        merged_lines = self.create_merged_line(segment[: i + j + 2])
                        segment[: i + j + 2] = merged_lines
                    except CannotMergeException:
                        pass
            except SqlfmtSegmentError:
                pass
        return segments

    def _maybe_merge_operators(
        self,
        segments: List[Segment],
        op_tiers: List[OperatorPrecedence],
    ) -> List[Segment]:
        """
        Tries to merge runs of segments that start with operators into previous
        segments. Operators have a priority that determines a sort of hierarchy;
        if we can't merge a whole run of operators, we increase the priority to
        create shorter runs that can be merged
        """
        if len(segments) <= 1 or not op_tiers:
            return segments
        head = 0
        new_segments: List[Segment] = []
        precedence = op_tiers.pop()

        for i, segment in enumerate(segments[1:], start=1):
            if not self._segment_continues_operator_sequence(segment, precedence):
                new_segments.extend(
                    self._try_merge_operator_segments(segments[head:i], op_tiers.copy())
                )
                head = i

        # we need to try one more time to merge everything after head
        else:
            new_segments.extend(
                self._try_merge_operator_segments(segments[head:], op_tiers.copy())
            )

        return new_segments

    @classmethod
    def _segment_continues_operator_sequence(
        cls, segment: Segment, max_precedence: OperatorPrecedence
    ) -> bool:
        """
        Returns true if the first line of the segment is part
        of a sequence of operators of priority <= max_priority
        """
        try:
            line, _ = segment.head
        except SqlfmtSegmentError:
            # if a segment is blank, keep scanning
            return True
        else:
            return (
                line.starts_with_operator
                and not line.previous_token_is_comma
                and OperatorPrecedence.from_node(line.nodes[0]) <= max_precedence
            ) or line.starts_with_comma

    def _try_merge_operator_segments(
        self, segments: List[Segment], op_tiers: List[OperatorPrecedence]
    ) -> List[Segment]:
        """
        Attempts to merge segments into a single line; if that fails,
        recurses at a lower operator priority
        """
        if len(segments) <= 1:
            return segments

        try:
            new_segments = [
                Segment(self.create_merged_line(list(itertools.chain(*segments))))
            ]
        except CannotMergeException:
            new_segments = self._maybe_merge_operators(segments, op_tiers)

        return new_segments

    def _maybe_stubbornly_merge(self, segments: List[Segment]) -> List[Segment]:
        """
        We prefer some operators, like `as`, `over()`, `exclude()`, and
        array or dictionary accessing with `[]` to be
        forced onto the prior line, even if the contents of their brackets
        don't fit there. This is also true for most operators that open
        a bracket, like `in ()` or `+ ()`, as long as the preceding segment
        does not also start with an operator.

        This method scans for segments that start with
        such operators and partially merges those segments with the prior
        segments by calling _stubbornly_merge()
        """
        if len(segments) <= 1:
            return segments

        new_segments = [segments[0]]

        # first stubborn-merge all p0 operators
        for segment in segments[1:]:
            if (
                # always stubbornly merge P0 operators (e.g., `over`)
                self._segment_continues_operator_sequence(
                    segment, max_precedence=OperatorPrecedence.OTHER_TIGHT
                )
            ):
                new_segments = self._stubbornly_merge(new_segments, segment)
            else:
                new_segments.append(segment)

        if len(new_segments) == 1:
            return new_segments

        # next, stubbon-merge qualifying p1 operators
        segments = new_segments
        new_segments = [segments[0]]

        starts_with_p1_operator = [
            self._segment_continues_operator_sequence(
                segment, max_precedence=OperatorPrecedence.COMPARATORS
            )
            for segment in segments
        ]
        for i, segment in enumerate(segments[1:], start=1):
            if (
                not starts_with_p1_operator[i - 1]
                and starts_with_p1_operator[i]
                and Segment(self.safe_create_merged_line(segment)).tail_closes_head
            ):
                new_segments = self._stubbornly_merge(new_segments, segment)
            else:
                new_segments.append(segment)

        return new_segments

    def _stubbornly_merge(
        self, prev_segments: List[Segment], segment: Segment
    ) -> List[Segment]:
        """
        Attempts several different methods of merging the last segment in
        new_segments and segment. Returns a list of segments that represent the
        best possible merger of those segments
        """
        new_segments = prev_segments.copy()
        prev_segment = new_segments.pop()
        try:
            head, i = segment.head
        except SqlfmtSegmentError:
            new_segments.extend([prev_segment, segment])
            return new_segments

        # try to merge the first line of this segment with the previous segment
        try:
            prev_segment = Segment(self.create_merged_line(prev_segment + [head]))
            prev_segment.extend(segment[i + 1 :])
            new_segments.append(prev_segment)
        except CannotMergeException:
            # try to add this segment to the last line of the previous segment
            last_line, k = prev_segment.tail
            try:
                new_last_lines = self.create_merged_line([last_line] + segment)
                prev_segment[-(k + 1) :] = new_last_lines
                new_segments.append(prev_segment)
            except CannotMergeException:
                # try to add just the first line of this segment to the last
                # line of the previous segment
                try:
                    new_last_lines = self.create_merged_line([last_line, head])
                    prev_segment[-(k + 1) :] = new_last_lines
                    prev_segment.extend(segment[i + 1 :])
                    new_segments.append(prev_segment)
                except CannotMergeException:
                    # give up and just return the original segments
                    new_segments.extend([prev_segment, segment])

        return new_segments



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import (
    EOL,
    JINJA_START,
    NEWLINE,
    SQL_COMMENT,
    SQL_QUOTED_EXP,
    group,
)
from sqlfmt.rules.jinja import JINJA
from sqlfmt.tokens import TokenType

ALWAYS = [
    Rule(
        name="fmt_off",
        priority=0,
        pattern=group(r"(--|#) ?fmt: ?off ?") + EOL,
        action=partial(actions.add_node_to_buffer, token_type=TokenType.FMT_OFF),
    ),
    Rule(
        name="fmt_on",
        priority=1,
        pattern=group(r"(--|#) ?fmt: ?on ?") + EOL,
        action=partial(actions.add_node_to_buffer, token_type=TokenType.FMT_ON),
    ),
    # These match just the start of jinja tags, which allows
    # the parser to deal with nesting in a more powerful way than
    # regex allows
    Rule(
        name="jinja_start",
        priority=120,
        pattern=JINJA_START,
        action=partial(actions.lex_ruleset, new_ruleset=JINJA),
    ),
    # we should never match the end of a jinja tag by itself
    Rule(
        name="jinja_end",
        priority=130,
        pattern=group(r"[#}%]\}"),
        action=actions.raise_sqlfmt_bracket_error,
    ),
    Rule(
        name="quoted_name",
        priority=200,
        pattern=SQL_QUOTED_EXP,
        action=partial(actions.add_node_to_buffer, token_type=TokenType.QUOTED_NAME),
    ),
    Rule(
        name="comment",
        priority=300,
        pattern=SQL_COMMENT,
        action=actions.add_comment_to_buffer,
    ),
    Rule(
        name="comment_start",
        priority=310,
        pattern=group(r"/\*"),
        action=partial(
            actions.handle_potentially_nested_tokens,
            start_name="comment_start",
            end_name="comment_end",
            token_type=TokenType.COMMENT,
        ),
    ),
    Rule(
        name="comment_end",
        priority=320,
        pattern=group(r"\*/"),
        action=actions.raise_sqlfmt_bracket_error,
    ),
    Rule(
        name="semicolon",
        priority=350,
        pattern=group(r";"),
        action=actions.handle_semicolon,
    ),
    Rule(
        name="newline",
        priority=9000,
        pattern=group(NEWLINE),
        action=actions.handle_newline,
    ),
]


CORE = [
    *ALWAYS,
    Rule(
        # see https://github.com/tconbeer/sqlfmt/issues/461
        name="pg_operator",
        priority=299,  # comments are 300
        pattern=group(r"#(>>?|-|#)"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.OPERATOR),
    ),
    Rule(
        # see https://spark.apache.org/docs/latest/sql-ref-literals.html#integral-literal-syntax
        name="spark_int_literals",
        priority=400,
        pattern=group(
            r"(\+|-)?\d+(l|s|y)",
        ),
        action=actions.handle_number,
    ),
    Rule(
        # the (bd|d|f) groups add support for Spark fractional literals
        # https://spark.apache.org/docs/latest/sql-ref-literals.html#fractional-literals-syntax
        name="number",
        priority=401,
        pattern=group(
            r"0x[0-9a-f]+",
            r"0o[0-7]+",
            r"0b[0-1]+",
            r"(\+|-)?\d+(_\d+)*(\.\d*(_\d+)*)?(e(\+|-)?\d+)?(bd|d|f)?",
            r"(\+|-)?\.\d+(_\d+)*(e(\+|-)?\d+)?(bd|d|f)?",
        ),
        action=actions.handle_number,
    ),
    Rule(
        name="star",
        priority=410,
        pattern=group(r"\*"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.STAR),
    ),
    Rule(
        name="double_colon",
        priority=420,
        pattern=group(r"::"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.DOUBLE_COLON),
    ),
    Rule(
        name="walrus",
        priority=421,
        pattern=group(r":="),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.OPERATOR),
    ),
    Rule(
        name="colon",
        priority=430,
        pattern=group(r":"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.COLON),
    ),
    Rule(
        name="comma",
        priority=440,
        pattern=group(r","),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.COMMA),
    ),
    Rule(
        name="dot",
        priority=450,
        pattern=group(r"\."),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.DOT),
    ),
    Rule(
        name="bracket_open",
        priority=500,
        pattern=group(
            r"\[",
            r"\(",
            r"\{",
            # bq/athena uses angle brackets for type definitions for compound types
            r"(array|map|table|struct)\s*<",
        ),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.BRACKET_OPEN),
    ),
    Rule(
        name="bracket_close",
        priority=510,
        pattern=group(
            r"\]",
            r"\)",
            r"\}",
        ),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.BRACKET_CLOSE),
    ),
    Rule(
        name="other_identifiers",
        priority=600,
        pattern=group(
            r"@\w+",  # stages
            r"metadata\$\w+",  # Snowflake metadata columns
            r"\$\d+",  # pg placeholders
            r"\$\w+",  # variables
            r"%(\([^%()]+\))?s",  # psycopg placeholders
            r"\?\d+",  # bun placeholders
        ),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.NAME),
    ),
    Rule(
        name="angle_bracket_close",
        priority=790,
        pattern=group(
            r">",
        ),
        action=actions.handle_closing_angle_bracket,
    ),
    Rule(
        name="operator",
        priority=800,
        pattern=group(
            r"\|\|?\/",  # square or cube root ||/
            r"~=",  # geo compare
            r"!?~{1,2}\*?",  # posix like/not like
            r"\?(=|!|<=|<!)",  # regex lookahead/behind
            r"\?(-\||\|\||-|\|)",  # regex lookahead/behind
            r"@-@",  # length operator
            r"@@@?",  # center point operator; also text match
            r"<->",  # distance operator
            r"<=>",  # null-safe equals (mysql)
            r"@>",  # contains
            r"<@",  # contained by
            r"<>",
            r"\|?>>=?",
            r"<<(=|\|)?",
            r"=>",
            r"->>?",  # json extraction
            r"\?(&|\|)",  # pg json contains keys
            r"&&",
            r"&<\|?",  # not extends
            r"\|?&>",  # not extends
            r"<\^",  # below
            r">\^",  # above
            r"\?#",  # intersect
            r"\|\|",
            r"-\|-",
            r"[*+?]?\?",  # regex greedy/non-greedy, also ?
            r"!!",  # negate text match
            r"%%",  # psycopg escaped mod operator
            r"[+\-*/%&|^=<>!]=?",  # singles
        ),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.OPERATOR),
    ),
    Rule(
        name="name",
        priority=5000,
        pattern=group(r"\w+"),
        action=partial(actions.add_node_to_buffer, token_type=TokenType.NAME),
    ),
]



import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.operator_precedence import OperatorPrecedence


def test_tiers_sorted() -> None:
    tiers = OperatorPrecedence.tiers()
    assert tiers == sorted(tiers)


@pytest.mark.parametrize(
    "source_string,expected_precedence",
    [
        ("::", OperatorPrecedence.DOUBLE_COLON),
        ("as", OperatorPrecedence.AS),
        ("over", OperatorPrecedence.OTHER_TIGHT),
        ("^", OperatorPrecedence.EXPONENT),
        ("/", OperatorPrecedence.MULTIPLICATION),
        ("%", OperatorPrecedence.MULTIPLICATION),
        ("%%", OperatorPrecedence.MULTIPLICATION),
        ("+", OperatorPrecedence.ADDITION),
        ("-", OperatorPrecedence.ADDITION),
        ("in", OperatorPrecedence.MEMBERSHIP),
        ("not in", OperatorPrecedence.MEMBERSHIP),
        ("like", OperatorPrecedence.MEMBERSHIP),
        ("not like", OperatorPrecedence.MEMBERSHIP),
        ("similar to", OperatorPrecedence.MEMBERSHIP),
        ("between", OperatorPrecedence.MEMBERSHIP),
        ("=", OperatorPrecedence.COMPARATORS),
        ("<=", OperatorPrecedence.COMPARATORS),
        (">=", OperatorPrecedence.COMPARATORS),
        ("<", OperatorPrecedence.COMPARATORS),
        (">", OperatorPrecedence.COMPARATORS),
        ("<>", OperatorPrecedence.COMPARATORS),
        ("!=", OperatorPrecedence.COMPARATORS),
        ("is", OperatorPrecedence.PRESENCE),
        ("notnull", OperatorPrecedence.PRESENCE),
        ("not", OperatorPrecedence.BOOL_NOT),
        ("and", OperatorPrecedence.BOOL_AND),
        ("or", OperatorPrecedence.BOOL_OR),
        ("||/", OperatorPrecedence.OTHER),
        ("<@", OperatorPrecedence.OTHER),
        ("exists", OperatorPrecedence.OTHER),
    ],
)
def test_operator_precedence(
    source_string: str,
    expected_precedence: OperatorPrecedence,
    default_analyzer: Analyzer,
) -> None:
    q = default_analyzer.parse_query(source_string)
    node = q.nodes[0]
    precedence = OperatorPrecedence.from_node(node)
    assert precedence == expected_precedence


@pytest.mark.parametrize(
    "source_string,expected_precedence",
    [
        ("1 *", OperatorPrecedence.MULTIPLICATION),
        ("my_array[", OperatorPrecedence.SQUARE_BRACKETS),
    ],
)
def test_operator_precedence_second_node(
    source_string: str,
    expected_precedence: OperatorPrecedence,
    default_analyzer: Analyzer,
) -> None:
    q = default_analyzer.parse_query(source_string)
    node = q.nodes[1]
    precedence = OperatorPrecedence.from_node(node)
    assert precedence == expected_precedence


def test_operator_precedence_between_and(default_analyzer: Analyzer) -> None:
    source_string = "foo between 1 and 3"
    q = default_analyzer.parse_query(source_string)
    and_node = q.nodes[3]
    assert OperatorPrecedence.from_node(and_node) == OperatorPrecedence.OTHER_TIGHT


@pytest.mark.parametrize("source_string", ["select", "my_table"])
def test_operator_precedence_raises(
    source_string: str, default_analyzer: Analyzer
) -> None:
    q = default_analyzer.parse_query(source_string)
    node = q.nodes[0]
    with pytest.raises(AssertionError):
        _ = OperatorPrecedence.from_node(node)



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import (
    JINJA_START,
    NEWLINE,
    SQL_COMMENT_START,
    SQL_QUOTED_EXP,
    group,
)
from sqlfmt.rules.core import ALWAYS
from sqlfmt.tokens import TokenType

UNSUPPORTED = [
    *ALWAYS,
    # quoted names need to be lexed as DATA, so they are not formatted.
    # quoted names are otherwise in ALWAYS, (so we don't lex them as comments etc.)
    # so this rule needs to have a higher priority than the quoted_name rule in
    # ALWAYS
    Rule(
        name="quoted_name_in_unsupported",
        priority=199,
        pattern=SQL_QUOTED_EXP,
        action=partial(actions.add_node_to_buffer, token_type=TokenType.DATA),
    ),
    Rule(
        name="unsupported_line",
        priority=1000,
        # a semicolon inside a string literal does not terminate the line, so
        # quoted expressions have to be consumed whole before we look for the
        # terminator
        # similarly, a semicolon inside a comment doesn't terminate an expression,
        # so we need to stop lexing at the start of a comment to let the higher-priority
        # comment rule consume that comment.
        pattern=group(rf"(?:{SQL_QUOTED_EXP}|[^;\n])+?")
        + group(r"\s*")
        + group(r";", NEWLINE, SQL_COMMENT_START, JINJA_START, r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(actions.add_node_to_buffer, token_type=TokenType.DATA),
        ),
    ),
]



import sys
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from sqlfmt.exception import SqlfmtConfigError
from sqlfmt.mode import Mode
from sqlfmt.report import STDIN_PATH

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


Config = Dict[str, Union[bool, int, List[str], str, Path]]


def load_config_file(files: List[Path], config_path: Optional[Path]) -> Config:
    """
    files is a list of resolved, absolute paths (like the ones passed from the
    Click CLI). This finds a pyproject.toml file in the common parent directory
    of files (or in the common parent's parents).

    If config_path is provided, searching for a config via files will be skipped
    entirely.
    """
    if config_path is None:
        common_parents = _get_common_parents(files)
        config_path = _find_config_file(common_parents)

    config = _load_config_from_path(config_path)
    return config


def _get_common_parents(files: List[Path]) -> List[Path]:
    """
    For a list of absolute paths, returns a Set of paths for all
    of the common parents of files
    """
    assert files, "Must provide a list of paths"
    common_parents: Set[Path] = set()
    for p in files:
        if p == STDIN_PATH:
            break
        else:
            assert p.is_absolute()
            parents = set(p.parents)
            if p.is_dir():
                parents.add(p)
            if not common_parents:
                common_parents = parents
            else:
                common_parents &= parents

    # the root directory is the lowest (i.e. most specific)
    # common parent among all of the files passed to sqlfmt
    try:
        root_dir = max(common_parents, key=lambda p: p.parts)
    except ValueError:
        # if there are no common parents (e.g., stdin), just use the cwd
        root_dir = Path.cwd()
    search_paths = [root_dir, *root_dir.parents]
    return search_paths


def _find_config_file(search_paths: List[Path]) -> Optional[Path]:
    """
    Given an ordered list of directories, returns the path to a
    pyproject.toml file in the lowest (most specific) directory

    Returns None if no file exists
    """
    for f in [p / "pyproject.toml" for p in search_paths]:
        if f.exists():
            return f
    else:
        return None


def _load_config_from_path(config_path: Optional[Path]) -> Config:
    """
    Loads a toml file located at config path. Returns the contents
    under the [tool.sqlfmt] section as a dict.
    """

    if not config_path:
        return {}
    else:
        try:
            with open(config_path, "rb") as f:
                pyproject_dict = tomllib.load(f)
        except OSError as e:
            raise SqlfmtConfigError(
                f"Error opening pyproject.toml config file at {config_path}. {e}"
            ) from e
        except tomllib.TOMLDecodeError as e:
            raise SqlfmtConfigError(
                f"Error decoding pyproject.toml config file at {config_path}. "
                f"Check for invalid TOML. {e}"
            ) from e
        raw_config: Config = pyproject_dict.get("tool", {}).get("sqlfmt", {})
        if "exclude" in raw_config and "exclude_root" not in raw_config:
            raw_config["exclude_root"] = config_path.parent
        return _validate_config(raw_config)


def _validate_config(raw_config: Config) -> Config:
    config = {}
    for k, v in ((k.lower(), v) for k, v in raw_config.items()):
        if k == "dialect":
            config["dialect_name"] = v
        elif k not in Mode.__dataclass_fields__:
            raise SqlfmtConfigError(
                f"Config file contains key {k}, which is not a "
                f"supported option. Must be one of "
                f"{list(Mode.__dataclass_fields__.keys())}"
            )
        else:
            config[k] = v
    return config



import re
from dataclasses import dataclass, field
from typing import List, Optional

from sqlfmt.comment import Comment
from sqlfmt.exception import SqlfmtBracketError, SqlfmtParsingError
from sqlfmt.line import Line
from sqlfmt.node import Node
from sqlfmt.node_manager import NodeManager
from sqlfmt.query import Query
from sqlfmt.rule import Rule


@dataclass
class Analyzer:
    """
    The Analyzer is initialized by the dialect specified by the user. The
    dialect specifies the list of rules that the Analyzer will attempt
    to match to the source string during parsing. The analyzer maintains
    buffers of lexed nodes, comments, and lines.
    """

    line_length: int
    rules: List[Rule]
    node_manager: NodeManager
    node_buffer: List[Node] = field(default_factory=list)
    comment_buffer: List[Comment] = field(default_factory=list)
    line_buffer: List[Line] = field(default_factory=list)
    rule_stack: List[List[Rule]] = field(default_factory=list)
    pos: int = 0

    @property
    def previous_node(self) -> Optional[Node]:
        """
        Return the most recently-lexed Node, if it exists
        """
        if self.node_buffer:
            return self.node_buffer[-1]
        else:
            return self.previous_line_node

    @property
    def previous_line_node(self) -> Optional[Node]:
        """
        Return the last node of the last complete line
        """
        if self.line_buffer:
            return self.line_buffer[-1].nodes[-1]
        else:
            return None

    def clear_buffers(self) -> None:
        """
        Reset the analyer's node, comment, and line buffers, and its parsing position.
        (It is possible for the same analyzer to be used to lex twice, so we need to
        reset buffers before lexing begins)
        """
        self.node_buffer = []
        self.comment_buffer = []
        self.line_buffer = []
        self.pos = 0

    def write_buffers_to_query(self, query: Query) -> None:
        """
        Write the contents of self.line_buffer to query.lines,
        taking care to flush node_buffer and comment_buffer first
        """
        # append a final line if the file doesn't end with a newline
        if self.node_buffer or self.comment_buffer:
            line = Line.from_nodes(
                previous_node=self.previous_line_node,
                nodes=self.node_buffer,
                comments=self.comment_buffer,
            )
            self.node_manager.append_newline(line)
            self.line_buffer.append(line)

        # if the final line(s) are jinja block end tags, they may be
        # indented too far -- they should be formatted as if they
        # have no open_brackets
        for line in reversed(self.line_buffer):
            if (
                line.is_standalone_jinja_statement
                and line.closes_jinja_block_from_previous_line
            ):
                for node in line.nodes:
                    node.open_brackets = []
            else:
                break
        query.lines = self.line_buffer

    def parse_query(self, source_string: str) -> Query:
        """
        Initialize a parser and parse the source string, return
        a structured Query.
        """
        q = Query(source_string, line_length=self.line_length)

        self.clear_buffers()
        self.lex(source_string=source_string)
        self.write_buffers_to_query(q)
        return q

    def push_rules(self, new_rules: List[Rule]) -> None:
        self.rule_stack.append(self.rules.copy())
        self.rules = sorted(new_rules, key=lambda rule: rule.priority)

    def pop_rules(self) -> List[Rule]:
        old_rules = self.rules
        self.rules = self.rule_stack.pop()
        return old_rules

    def get_rule(self, rule_name: str) -> Rule:
        """
        Return the rule from ruleset that matches rule_name
        """
        matching_rules = filter(lambda rule: rule.name == rule_name, self.rules)
        try:
            return next(matching_rules)
        except StopIteration as e:
            raise ValueError(f"No rule '{rule_name}'") from e

    def lex_one(self, source_string: str) -> None:
        """
        Repeatedly match Rules to the source_string (at self.pos)
        and apply the matched action.

        Mutates the analyzer's buffers and pos
        """
        for rule in self.rules:
            match = rule.program.match(source_string, self.pos)
            if match:
                rule.action(self, source_string, match)
                return
        # nothing matched. Either whitespace or an error
        else:
            raise SqlfmtParsingError(
                f"Could not parse SQL at position {self.pos}:"
                f" '{source_string[self.pos : self.pos + 50].strip()}'"
            )

    def lex(self, source_string: str, eof_pos: int = -1) -> None:
        """
        Repeatedly match Rules to the source_string (until the source_string is
        exhausted) and apply the matched action.

        Mutates the analyzer's buffers
        """
        if eof_pos == -1:
            for idx, char in enumerate(reversed(source_string)):
                if not char.isspace():
                    eof_pos = len(source_string) - idx
                    break

        last_loop_pos = -1
        while self.pos < eof_pos and self.pos > last_loop_pos:
            last_loop_pos = self.pos
            self.lex_one(source_string)

    def search_for_terminating_token(
        self,
        start_rule: str,
        program: re.Pattern,
        nesting_program: re.Pattern,
        tail: str,
        pos: int,
    ) -> int:
        """
        Return the ending position of the correct closing bracket that matches
        start_rule
        """

        match = program.search(tail)
        if not match:
            raise SqlfmtBracketError(
                f"Unterminated multiline token '{start_rule}' "
                f"started near position {pos}."
            )

        start, end = match.span(1)

        nesting_match = nesting_program.match(tail, start)
        if nesting_match:
            inner_epos = self.search_for_terminating_token(
                start_rule=start_rule,
                program=program,
                nesting_program=nesting_program,
                tail=tail[end:],
                pos=pos + end,
            )
            outer_epos = self.search_for_terminating_token(
                start_rule=start_rule,
                program=program,
                nesting_program=nesting_program,
                tail=tail[inner_epos - pos :],
                pos=inner_epos,
            )
            return outer_epos
        else:
            return pos + end
